##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import unittest

class Test_parse_http_date(unittest.TestCase):

    def _callFUT(self, v):
        from waitress.utilities import parse_http_date
        return parse_http_date(v)

    def test_rfc850(self):
        val = 'Tuesday, 08-Feb-94 14:15:29 GMT'
        result = self._callFUT(val)
        self.assertEqual(result, 760716929)

    def test_rfc822(self):
        val = 'Sun, 08 Feb 1994 14:15:29 GMT'
        result = self._callFUT(val)
        self.assertEqual(result, 760716929)

    def test_neither(self):
        val = ''
        result = self._callFUT(val)
        self.assertEqual(result, 0)

class Test_build_http_date(unittest.TestCase):

    def test_rountdrip(self):
        from waitress.utilities import build_http_date, parse_http_date
        from time import time
        t = int(time())
        self.assertEqual(t, parse_http_date(build_http_date(t)))

class Test_unpack_rfc850(unittest.TestCase):

    def _callFUT(self, val):
        from waitress.utilities import unpack_rfc850, rfc850_reg
        return unpack_rfc850(rfc850_reg.match(val.lower()))

    def test_it(self):
        val = 'Tuesday, 08-Feb-94 14:15:29 GMT'
        result = self._callFUT(val)
        self.assertEqual(result, (1994, 2, 8, 14, 15, 29, 0, 0, 0))

class Test_unpack_rfc_822(unittest.TestCase):

    def _callFUT(self, val):
        from waitress.utilities import unpack_rfc822, rfc822_reg
        return unpack_rfc822(rfc822_reg.match(val.lower()))

    def test_it(self):
        val = 'Sun, 08 Feb 1994 14:15:29 GMT'
        result = self._callFUT(val)
        self.assertEqual(result, (1994, 2, 8, 14, 15, 29, 0, 0, 0))

class Test_find_double_newline(unittest.TestCase):

    def _callFUT(self, val):
        from waitress.utilities import find_double_newline
        return find_double_newline(val)

    def test_empty(self):
        self.assertEqual(self._callFUT(b''), -1)

    def test_one_linefeed(self):
        self.assertEqual(self._callFUT(b'\n'), -1)

    def test_double_linefeed(self):
        self.assertEqual(self._callFUT(b'\n\n'), 2)

    def test_one_crlf(self):
        self.assertEqual(self._callFUT(b'\r\n'), -1)

    def test_double_crfl(self):
        self.assertEqual(self._callFUT(b'\r\n\r\n'), 4)

    def test_mixed(self):
        self.assertEqual(self._callFUT(b'\n\n00\r\n\r\n'), 2)

class TestBadRequest(unittest.TestCase):

    def _makeOne(self):
        from waitress.utilities import BadRequest
        return BadRequest(1)

    def test_it(self):
        inst = self._makeOne()
        self.assertEqual(inst.body, 1)

class Test_undquote(unittest.TestCase):

    def _callFUT(self, value):
        from waitress.utilities import undquote
        return undquote(value)

    def test_empty(self):
        self.assertEqual(self._callFUT(''), '')

    def test_quoted(self):
        self.assertEqual(self._callFUT('"test"'), 'test')

    def test_unquoted(self):
        self.assertEqual(self._callFUT('test'), 'test')

    def test_quoted_backslash_quote(self):
        self.assertEqual(self._callFUT('"\\""'), '"')

    def test_quoted_htab(self):
        self.assertEqual(self._callFUT("\"\t\""), "\t")

    def test_quoted_backslash_htab(self):
        self.assertEqual(self._callFUT("\"\\\t\""), "\t")

    def test_quoted_backslash_invalid(self):
        self.assertRaises(ValueError, self._callFUT, '"\\"')

    def test_invalid_quoting(self):
        self.assertRaises(ValueError, self._callFUT, '"test')

    def test_invalid_quoting_single_quote(self):
        self.assertRaises(ValueError, self._callFUT, '"')




##############################################################################
#
# Copyright (c) 2001-2004 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Buffers
"""
from io import BytesIO

# copy_bytes controls the size of temp. strings for shuffling data around.
COPY_BYTES = 1 << 18 # 256K

# The maximum number of bytes to buffer in a simple string.
STRBUF_LIMIT = 8192

class FileBasedBuffer(object):

    remain = 0

    def __init__(self, file, from_buffer=None):
        self.file = file
        if from_buffer is not None:
            from_file = from_buffer.getfile()
            read_pos = from_file.tell()
            from_file.seek(0)
            while True:
                data = from_file.read(COPY_BYTES)
                if not data:
                    break
                file.write(data)
            self.remain = int(file.tell() - read_pos)
            from_file.seek(read_pos)
            file.seek(read_pos)

    def __len__(self):
        return self.remain

    def __nonzero__(self):
        return True

    __bool__ = __nonzero__ # py3

    def append(self, s):
        file = self.file
        read_pos = file.tell()
        file.seek(0, 2)
        file.write(s)
        file.seek(read_pos)
        self.remain = self.remain + len(s)

    def get(self, numbytes=-1, skip=False):
        file = self.file
        if not skip:
            read_pos = file.tell()
        if numbytes < 0:
            # Read all
            res = file.read()
        else:
            res = file.read(numbytes)
        if skip:
            self.remain -= len(res)
        else:
            file.seek(read_pos)
        return res

    def skip(self, numbytes, allow_prune=0):
        if self.remain < numbytes:
            raise ValueError("Can't skip %d bytes in buffer of %d bytes" % (
                numbytes, self.remain)
            )
        self.file.seek(numbytes, 1)
        self.remain = self.remain - numbytes

    def newfile(self):
        raise NotImplementedError()

    def prune(self):
        file = self.file
        if self.remain == 0:
            read_pos = file.tell()
            file.seek(0, 2)
            sz = file.tell()
            file.seek(read_pos)
            if sz == 0:
                # Nothing to prune.
                return
        nf = self.newfile()
        while True:
            data = file.read(COPY_BYTES)
            if not data:
                break
            nf.write(data)
        self.file = nf

    def getfile(self):
        return self.file

    def close(self):
        if hasattr(self.file, 'close'):
            self.file.close()
        self.remain = 0

class TempfileBasedBuffer(FileBasedBuffer):

    def __init__(self, from_buffer=None):
        FileBasedBuffer.__init__(self, self.newfile(), from_buffer)

    def newfile(self):
        from tempfile import TemporaryFile
        return TemporaryFile('w+b')

class BytesIOBasedBuffer(FileBasedBuffer):

    def __init__(self, from_buffer=None):
        if from_buffer is not None:
            FileBasedBuffer.__init__(self, BytesIO(), from_buffer)
        else:
            # Shortcut. :-)
            self.file = BytesIO()

    def newfile(self):
        return BytesIO()

class ReadOnlyFileBasedBuffer(FileBasedBuffer):
    # used as wsgi.file_wrapper

    def __init__(self, file, block_size=32768):
        self.file = file
        self.block_size = block_size # for __iter__

    def prepare(self, size=None):
        if hasattr(self.file, 'seek') and hasattr(self.file, 'tell'):
            start_pos = self.file.tell()
            self.file.seek(0, 2)
            end_pos = self.file.tell()
            self.file.seek(start_pos)
            fsize = end_pos - start_pos
            if size is None:
                self.remain = fsize
            else:
                self.remain = min(fsize, size)
        return self.remain

    def get(self, numbytes=-1, skip=False):
        # never read more than self.remain (it can be user-specified)
        if numbytes == -1 or numbytes > self.remain:
            numbytes = self.remain
        file = self.file
        if not skip:
            read_pos = file.tell()
        res = file.read(numbytes)
        if skip:
            self.remain -= len(res)
        else:
            file.seek(read_pos)
        return res

    def __iter__(self): # called by task if self.filelike has no seek/tell
        return self

    def next(self):
        val = self.file.read(self.block_size)
        if not val:
            raise StopIteration
        return val

    __next__ = next # py3

    def append(self, s):
        raise NotImplementedError

class OverflowableBuffer(object):
    """
    This buffer implementation has four stages:
    - No data
    - Bytes-based buffer
    - BytesIO-based buffer
    - Temporary file storage
    The first two stages are fastest for simple transfers.
    """

    overflowed = False
    buf = None
    strbuf = b'' # Bytes-based buffer.

    def __init__(self, overflow):
        # overflow is the maximum to be stored in a StringIO buffer.
        self.overflow = overflow

    def __len__(self):
        buf = self.buf
        if buf is not None:
            # use buf.__len__ rather than len(buf) FBO of not getting
            # OverflowError on Python 2
            return buf.__len__()
        else:
            return self.strbuf.__len__()

    def __nonzero__(self):
        # use self.__len__ rather than len(self) FBO of not getting
        # OverflowError on Python 2
        return self.__len__() > 0

    __bool__ = __nonzero__ # py3

    def _create_buffer(self):
        strbuf = self.strbuf
        if len(strbuf) >= self.overflow:
            self._set_large_buffer()
        else:
            self._set_small_buffer()
        buf = self.buf
        if strbuf:
            buf.append(self.strbuf)
            self.strbuf = b''
        return buf

    def _set_small_buffer(self):
        self.buf = BytesIOBasedBuffer(self.buf)
        self.overflowed = False

    def _set_large_buffer(self):
        self.buf = TempfileBasedBuffer(self.buf)
        self.overflowed = True

    def append(self, s):
        buf = self.buf
        if buf is None:
            strbuf = self.strbuf
            if len(strbuf) + len(s) < STRBUF_LIMIT:
                self.strbuf = strbuf + s
                return
            buf = self._create_buffer()
        buf.append(s)
        # use buf.__len__ rather than len(buf) FBO of not getting
        # OverflowError on Python 2
        sz = buf.__len__()
        if not self.overflowed:
            if sz >= self.overflow:
                self._set_large_buffer()

    def get(self, numbytes=-1, skip=False):
        buf = self.buf
        if buf is None:
            strbuf = self.strbuf
            if not skip:
                return strbuf
            buf = self._create_buffer()
        return buf.get(numbytes, skip)

    def skip(self, numbytes, allow_prune=False):
        buf = self.buf
        if buf is None:
            if allow_prune and numbytes == len(self.strbuf):
                # We could slice instead of converting to
                # a buffer, but that would eat up memory in
                # large transfers.
                self.strbuf = b''
                return
            buf = self._create_buffer()
        buf.skip(numbytes, allow_prune)

    def prune(self):
        """
        A potentially expensive operation that removes all data
        already retrieved from the buffer.
        """
        buf = self.buf
        if buf is None:
            self.strbuf = b''
            return
        buf.prune()
        if self.overflowed:
            # use buf.__len__ rather than len(buf) FBO of not getting
            # OverflowError on Python 2
            sz = buf.__len__()
            if sz < self.overflow:
                # Revert to a faster buffer.
                self._set_small_buffer()

    def getfile(self):
        buf = self.buf
        if buf is None:
            buf = self._create_buffer()
        return buf.getfile()

    def close(self):
        buf = self.buf
        if buf is not None:
            buf.close()



import sys

if __name__ == '__main__':
    try:
        from urllib.request import urlopen, URLError
    except ImportError:
        from urllib2 import urlopen, URLError

    url = sys.argv[1]
    headers = {'Content-Type': 'text/plain; charset=utf-8'}
    try:
        resp = urlopen(url)
        line = resp.readline().decode('ascii') # py3
    except URLError:
        line = 'failed to read %s' % url
    sys.stdout.write(line)
    sys.stdout.flush()



import time

def app(environ, start_response): # pragma: no cover
    if environ['PATH_INFO'] == '/sleepy':
        time.sleep(2)
        body = b'sleepy returned'
    else:
        body = b'notsleepy returned'
    cl = str(len(body))
    start_response(
        '200 OK',
        [('Content-Length', cl), ('Content-Type', 'text/plain')]
    )
    return [body]



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""HTTP Request Parser

This server uses asyncore to accept connections and do initial
processing but threads to do work.
"""
import re
from io import BytesIO

from waitress.compat import (
    tostr,
    urlparse,
    unquote_bytes_to_wsgi,
)

from waitress.buffers import OverflowableBuffer

from waitress.receiver import (
    FixedStreamReceiver,
    ChunkedReceiver,
)

from waitress.utilities import (
    find_double_newline,
    RequestEntityTooLarge,
    RequestHeaderFieldsTooLarge,
    BadRequest,
)

class ParsingError(Exception):
    pass

class HTTPRequestParser(object):
    """A structure that collects the HTTP request.

    Once the stream is completed, the instance is passed to
    a server task constructor.
    """
    completed = False        # Set once request is completed.
    empty = False            # Set if no request was made.
    expect_continue = False  # client sent "Expect: 100-continue" header
    headers_finished = False # True when headers have been read
    header_plus = b''
    chunked = False
    content_length = 0
    header_bytes_received = 0
    body_bytes_received = 0
    body_rcv = None
    version = '1.0'
    error = None
    connection_close = False

    # Other attributes: first_line, header, headers, command, uri, version,
    # path, query, fragment

    def __init__(self, adj):
        """
        adj is an Adjustments object.
        """
        # headers is a mapping containing keys translated to uppercase
        # with dashes turned into underscores.
        self.headers = {}
        self.adj = adj

    def received(self, data):
        """
        Receives the HTTP stream for one request.  Returns the number of
        bytes consumed.  Sets the completed flag once both the header and the
        body have been received.
        """
        if self.completed:
            return 0 # Can't consume any more.
        datalen = len(data)
        br = self.body_rcv
        if br is None:
            # In header.
            s = self.header_plus + data
            index = find_double_newline(s)
            if index >= 0:
                # Header finished.
                header_plus = s[:index]
                consumed = len(data) - (len(s) - index)
                # Remove preceeding blank lines.
                header_plus = header_plus.lstrip()
                if not header_plus:
                    self.empty = True
                    self.completed = True
                else:
                    try:
                        self.parse_header(header_plus)
                    except ParsingError as e:
                        self.error = BadRequest(e.args[0])
                        self.completed = True
                    else:
                        if self.body_rcv is None:
                            # no content-length header and not a t-e: chunked
                            # request
                            self.completed = True
                        if self.content_length > 0:
                            max_body = self.adj.max_request_body_size
                            # we won't accept this request if the content-length
                            # is too large
                            if self.content_length >= max_body:
                                self.error = RequestEntityTooLarge(
                                    'exceeds max_body of %s' % max_body)
                                self.completed = True
                self.headers_finished = True
                return consumed
            else:
                # Header not finished yet.
                self.header_bytes_received += datalen
                max_header = self.adj.max_request_header_size
                if self.header_bytes_received >= max_header:
                    # malformed header, we need to construct some request
                    # on our own. we disregard the incoming(?) requests HTTP
                    # version and just use 1.0. IOW someone just sent garbage
                    # over the wire
                    self.parse_header(b'GET / HTTP/1.0\n')
                    self.error = RequestHeaderFieldsTooLarge(
                        'exceeds max_header of %s' % max_header)
                    self.completed = True
                self.header_plus = s
                return datalen
        else:
            # In body.
            consumed = br.received(data)
            self.body_bytes_received += consumed
            max_body = self.adj.max_request_body_size
            if self.body_bytes_received >= max_body:
                # this will only be raised during t-e: chunked requests
                self.error = RequestEntityTooLarge(
                    'exceeds max_body of %s' % max_body)
                self.completed = True
            elif br.error:
                # garbage in chunked encoding input probably
                self.error = br.error
                self.completed = True
            elif br.completed:
                # The request (with the body) is ready to use.
                self.completed = True
                if self.chunked:
                    # We've converted the chunked transfer encoding request
                    # body into a normal request body, so we know its content
                    # length; set the header here.  We already popped the
                    # TRANSFER_ENCODING header in parse_header, so this will
                    # appear to the client to be an entirely non-chunked HTTP
                    # request with a valid content-length.
                    self.headers['CONTENT_LENGTH'] = str(br.__len__())
            return consumed

    def parse_header(self, header_plus):
        """
        Parses the header_plus block of text (the headers plus the
        first line of the request).
        """
        index = header_plus.find(b'\n')
        if index >= 0:
            first_line = header_plus[:index].rstrip()
            header = header_plus[index + 1:]
        else:
            first_line = header_plus.rstrip()
            header = b''

        self.first_line = first_line # for testing

        lines = get_header_lines(header)

        headers = self.headers
        for line in lines:
            index = line.find(b':')
            if index > 0:
                key = line[:index]
                if b'_' in key:
                    continue
                value = line[index + 1:].strip()
                key1 = tostr(key.upper().replace(b'-', b'_'))
                # If a header already exists, we append subsequent values
                # seperated by a comma. Applications already need to handle
                # the comma seperated values, as HTTP front ends might do
                # the concatenation for you (behavior specified in RFC2616).
                try:
                    headers[key1] += tostr(b', ' + value)
                except KeyError:
                    headers[key1] = tostr(value)
            # else there's garbage in the headers?

        # command, uri, version will be bytes
        command, uri, version = crack_first_line(first_line)
        version = tostr(version)
        command = tostr(command)
        self.command = command
        self.version = version
        (self.proxy_scheme,
         self.proxy_netloc,
         self.path,
         self.query, self.fragment) = split_uri(uri)
        self.url_scheme = self.adj.url_scheme
        connection = headers.get('CONNECTION', '')

        if version == '1.0':
            if connection.lower() != 'keep-alive':
                self.connection_close = True

        if version == '1.1':
            # since the server buffers data from chunked transfers and clients
            # never need to deal with chunked requests, downstream clients
            # should not see the HTTP_TRANSFER_ENCODING header; we pop it
            # here
            te = headers.pop('TRANSFER_ENCODING', '')
            if te.lower() == 'chunked':
                self.chunked = True
                buf = OverflowableBuffer(self.adj.inbuf_overflow)
                self.body_rcv = ChunkedReceiver(buf)
            expect = headers.get('EXPECT', '').lower()
            self.expect_continue = expect == '100-continue'
            if connection.lower() == 'close':
                self.connection_close = True

        if not self.chunked:
            try:
                cl = int(headers.get('CONTENT_LENGTH', 0))
            except ValueError:
                cl = 0
            self.content_length = cl
            if cl > 0:
                buf = OverflowableBuffer(self.adj.inbuf_overflow)
                self.body_rcv = FixedStreamReceiver(cl, buf)

    def get_body_stream(self):
        body_rcv = self.body_rcv
        if body_rcv is not None:
            return body_rcv.getfile()
        else:
            return BytesIO()

    def close(self):
        body_rcv = self.body_rcv
        if body_rcv is not None:
            body_rcv.getbuf().close()

def split_uri(uri):
    # urlsplit handles byte input by returning bytes on py3, so
    # scheme, netloc, path, query, and fragment are bytes
    try:
        scheme, netloc, path, query, fragment = urlparse.urlsplit(uri)
    except UnicodeError:
        raise ParsingError('Bad URI')
    return (
        tostr(scheme),
        tostr(netloc),
        unquote_bytes_to_wsgi(path),
        tostr(query),
        tostr(fragment),
    )

def get_header_lines(header):
    """
    Splits the header into lines, putting multi-line headers together.
    """
    r = []
    lines = header.split(b'\n')
    for line in lines:
        if line.startswith((b' ', b'\t')):
            if not r:
                # https://corte.si/posts/code/pathod/pythonservers/index.html
                raise ParsingError('Malformed header line "%s"' % tostr(line))
            r[-1] += line
        else:
            r.append(line)
    return r

first_line_re = re.compile(
    b'([^ ]+) '
    b'((?:[^ :?#]+://[^ ?#/]*(?:[0-9]{1,5})?)?[^ ]+)'
    b'(( HTTP/([0-9.]+))$|$)'
)

def crack_first_line(line):
    m = first_line_re.match(line)
    if m is not None and m.end() == len(line):
        if m.group(3):
            version = m.group(5)
        else:
            version = None
        method = m.group(1)

        # the request methods that are currently defined are all uppercase:
        # https://www.iana.org/assignments/http-methods/http-methods.xhtml and
        # the request method is case sensitive according to
        # https://tools.ietf.org/html/rfc7231#section-4.1

        # By disallowing anything but uppercase methods we save poor
        # unsuspecting souls from sending lowercase HTTP methods to waitress
        # and having the request complete, while servers like nginx drop the
        # request onto the floor.
        if method != method.upper():
            raise ParsingError('Malformed HTTP method "%s"' % tostr(method))
        uri = m.group(2)
        return method, uri, version
    else:
        return b'', b'', b''



import unittest

class Test_serve(unittest.TestCase):

    def _callFUT(self, app, **kw):
        from waitress import serve
        return serve(app, **kw)

    def test_it(self):
        server = DummyServerFactory()
        app = object()
        result = self._callFUT(app, _server=server, _quiet=True)
        self.assertEqual(server.app, app)
        self.assertEqual(result, None)
        self.assertEqual(server.ran, True)


class Test_serve_paste(unittest.TestCase):

    def _callFUT(self, app, **kw):
        from waitress import serve_paste
        return serve_paste(app, None, **kw)

    def test_it(self):
        server = DummyServerFactory()
        app = object()
        result = self._callFUT(app, _server=server, _quiet=True)
        self.assertEqual(server.app, app)
        self.assertEqual(result, 0)
        self.assertEqual(server.ran, True)

class DummyServerFactory(object):
    ran = False

    def __call__(self, app, **kw):
        self.adj = DummyAdj(kw)
        self.app = app
        self.kw = kw
        return self

    def run(self):
        self.ran = True

class DummyAdj(object):
    verbose = False

    def __init__(self, kw):
        self.__dict__.update(kw)



import unittest
import io

class TestThreadedTaskDispatcher(unittest.TestCase):

    def _makeOne(self):
        from waitress.task import ThreadedTaskDispatcher
        return ThreadedTaskDispatcher()

    def test_handler_thread_task_is_None(self):
        inst = self._makeOne()
        inst.threads[0] = True
        inst.queue.put(None)
        inst.handler_thread(0)
        self.assertEqual(inst.stop_count, -1)
        self.assertEqual(inst.threads, {})

    def test_handler_thread_task_raises(self):
        from waitress.task import JustTesting
        inst = self._makeOne()
        inst.threads[0] = True
        inst.logger = DummyLogger()
        task = DummyTask(JustTesting)
        inst.logger = DummyLogger()
        inst.queue.put(task)
        inst.handler_thread(0)
        self.assertEqual(inst.stop_count, -1)
        self.assertEqual(inst.threads, {})
        self.assertEqual(len(inst.logger.logged), 1)

    def test_set_thread_count_increase(self):
        inst = self._makeOne()
        L = []
        inst.start_new_thread = lambda *x: L.append(x)
        inst.set_thread_count(1)
        self.assertEqual(L, [(inst.handler_thread, (0,))])

    def test_set_thread_count_increase_with_existing(self):
        inst = self._makeOne()
        L = []
        inst.threads = {0: 1}
        inst.start_new_thread = lambda *x: L.append(x)
        inst.set_thread_count(2)
        self.assertEqual(L, [(inst.handler_thread, (1,))])

    def test_set_thread_count_decrease(self):
        inst = self._makeOne()
        inst.threads = {'a': 1, 'b': 2}
        inst.set_thread_count(1)
        self.assertEqual(inst.queue.qsize(), 1)
        self.assertEqual(inst.queue.get(), None)

    def test_set_thread_count_same(self):
        inst = self._makeOne()
        L = []
        inst.start_new_thread = lambda *x: L.append(x)
        inst.threads = {0: 1}
        inst.set_thread_count(1)
        self.assertEqual(L, [])

    def test_add_task(self):
        task = DummyTask()
        inst = self._makeOne()
        inst.add_task(task)
        self.assertEqual(inst.queue.qsize(), 1)
        self.assertTrue(task.deferred)

    def test_log_queue_depth(self):
        task = DummyTask()
        inst = self._makeOne()
        inst.queue_logger = DummyLogger()
        inst.add_task(task)
        self.assertEqual(len(inst.queue_logger.logged), 0)
        inst.add_task(task)
        self.assertEqual(len(inst.queue_logger.logged), 1)

    def test_add_task_defer_raises(self):
        task = DummyTask(ValueError)
        inst = self._makeOne()
        self.assertRaises(ValueError, inst.add_task, task)
        self.assertEqual(inst.queue.qsize(), 0)
        self.assertTrue(task.deferred)
        self.assertTrue(task.cancelled)

    def test_shutdown_one_thread(self):
        inst = self._makeOne()
        inst.threads[0] = 1
        inst.logger = DummyLogger()
        task = DummyTask()
        inst.queue.put(task)
        self.assertEqual(inst.shutdown(timeout=.01), True)
        self.assertEqual(inst.logger.logged, ['1 thread(s) still running'])
        self.assertEqual(task.cancelled, True)

    def test_shutdown_no_threads(self):
        inst = self._makeOne()
        self.assertEqual(inst.shutdown(timeout=.01), True)

    def test_shutdown_no_cancel_pending(self):
        inst = self._makeOne()
        self.assertEqual(inst.shutdown(cancel_pending=False, timeout=.01),
                         False)

class TestTask(unittest.TestCase):

    def _makeOne(self, channel=None, request=None):
        if channel is None:
            channel = DummyChannel()
        if request is None:
            request = DummyParser()
        from waitress.task import Task
        return Task(channel, request)

    def test_ctor_version_not_in_known(self):
        request = DummyParser()
        request.version = '8.4'
        inst = self._makeOne(request=request)
        self.assertEqual(inst.version, '1.0')

    def test_cancel(self):
        inst = self._makeOne()
        inst.cancel()
        self.assertTrue(inst.close_on_finish)

    def test_defer(self):
        inst = self._makeOne()
        self.assertEqual(inst.defer(), None)

    def test_build_response_header_bad_http_version(self):
        inst = self._makeOne()
        inst.request = DummyParser()
        inst.version = '8.4'
        self.assertRaises(AssertionError, inst.build_response_header)

    def test_build_response_header_v10_keepalive_no_content_length(self):
        inst = self._makeOne()
        inst.request = DummyParser()
        inst.request.headers['CONNECTION'] = 'keep-alive'
        inst.version = '1.0'
        result = inst.build_response_header()
        lines = filter_lines(result)
        self.assertEqual(len(lines), 4)
        self.assertEqual(lines[0], b'HTTP/1.0 200 OK')
        self.assertEqual(lines[1], b'Connection: close')
        self.assertTrue(lines[2].startswith(b'Date:'))
        self.assertEqual(lines[3], b'Server: waitress')
        self.assertEqual(inst.close_on_finish, True)
        self.assertTrue(('Connection', 'close') in inst.response_headers)

    def test_build_response_header_v10_keepalive_with_content_length(self):
        inst = self._makeOne()
        inst.request = DummyParser()
        inst.request.headers['CONNECTION'] = 'keep-alive'
        inst.response_headers = [('Content-Length', '10')]
        inst.version = '1.0'
        inst.content_length = 0
        result = inst.build_response_header()
        lines = filter_lines(result)
        self.assertEqual(len(lines), 5)
        self.assertEqual(lines[0], b'HTTP/1.0 200 OK')
        self.assertEqual(lines[1], b'Connection: Keep-Alive')
        self.assertEqual(lines[2], b'Content-Length: 10')
        self.assertTrue(lines[3].startswith(b'Date:'))
        self.assertEqual(lines[4], b'Server: waitress')
        self.assertEqual(inst.close_on_finish, False)

    def test_build_response_header_v11_connection_closed_by_client(self):
        inst = self._makeOne()
        inst.request = DummyParser()
        inst.version = '1.1'
        inst.request.headers['CONNECTION'] = 'close'
        result = inst.build_response_header()
        lines = filter_lines(result)
        self.assertEqual(len(lines), 5)
        self.assertEqual(lines[0], b'HTTP/1.1 200 OK')
        self.assertEqual(lines[1], b'Connection: close')
        self.assertTrue(lines[2].startswith(b'Date:'))
        self.assertEqual(lines[3], b'Server: waitress')
        self.assertEqual(lines[4], b'Transfer-Encoding: chunked')
        self.assertTrue(('Connection', 'close') in inst.response_headers)
        self.assertEqual(inst.close_on_finish, True)

    def test_build_response_header_v11_connection_keepalive_by_client(self):
        inst = self._makeOne()
        inst.request = DummyParser()
        inst.request.headers['CONNECTION'] = 'keep-alive'
        inst.version = '1.1'
        result = inst.build_response_header()
        lines = filter_lines(result)
        self.assertEqual(len(lines), 5)
        self.assertEqual(lines[0], b'HTTP/1.1 200 OK')
        self.assertEqual(lines[1], b'Connection: close')
        self.assertTrue(lines[2].startswith(b'Date:'))
        self.assertEqual(lines[3], b'Server: waitress')
        self.assertEqual(lines[4], b'Transfer-Encoding: chunked')
        self.assertTrue(('Connection', 'close') in inst.response_headers)
        self.assertEqual(inst.close_on_finish, True)

    def test_build_response_header_v11_200_no_content_length(self):
        inst = self._makeOne()
        inst.request = DummyParser()
        inst.version = '1.1'
        result = inst.build_response_header()
        lines = filter_lines(result)
        self.assertEqual(len(lines), 5)
        self.assertEqual(lines[0], b'HTTP/1.1 200 OK')
        self.assertEqual(lines[1], b'Connection: close')
        self.assertTrue(lines[2].startswith(b'Date:'))
        self.assertEqual(lines[3], b'Server: waitress')
        self.assertEqual(lines[4], b'Transfer-Encoding: chunked')
        self.assertEqual(inst.close_on_finish, True)
        self.assertTrue(('Connection', 'close') in inst.response_headers)

    def test_build_response_header_v11_204_no_content_length_or_transfer_encoding(self):
        # RFC 7230: MUST NOT send Transfer-Encoding or Content-Length
        # for any response with a status code of 1xx or 204.
        inst = self._makeOne()
        inst.request = DummyParser()
        inst.version = '1.1'
        inst.status = '204 No Content'
        result = inst.build_response_header()
        lines = filter_lines(result)
        self.assertEqual(len(lines), 4)
        self.assertEqual(lines[0], b'HTTP/1.1 204 No Content')
        self.assertEqual(lines[1], b'Connection: close')
        self.assertTrue(lines[2].startswith(b'Date:'))
        self.assertEqual(lines[3], b'Server: waitress')
        self.assertEqual(inst.close_on_finish, True)
        self.assertTrue(('Connection', 'close') in inst.response_headers)

    def test_build_response_header_v11_1xx_no_content_length_or_transfer_encoding(self):
        # RFC 7230: MUST NOT send Transfer-Encoding or Content-Length
        # for any response with a status code of 1xx or 204.
        inst = self._makeOne()
        inst.request = DummyParser()
        inst.version = '1.1'
        inst.status = '100 Continue'
        result = inst.build_response_header()
        lines = filter_lines(result)
        self.assertEqual(len(lines), 4)
        self.assertEqual(lines[0], b'HTTP/1.1 100 Continue')
        self.assertEqual(lines[1], b'Connection: close')
        self.assertTrue(lines[2].startswith(b'Date:'))
        self.assertEqual(lines[3], b'Server: waitress')
        self.assertEqual(inst.close_on_finish, True)
        self.assertTrue(('Connection', 'close') in inst.response_headers)

    def test_build_response_header_v11_304_no_content_length_or_transfer_encoding(self):
        # RFC 7230: MUST NOT send Transfer-Encoding or Content-Length
        # for any response with a status code of 1xx, 204 or 304.
        inst = self._makeOne()
        inst.request = DummyParser()
        inst.version = '1.1'
        inst.status = '304 Not Modified'
        result = inst.build_response_header()
        lines = filter_lines(result)
        self.assertEqual(len(lines), 4)
        self.assertEqual(lines[0], b'HTTP/1.1 304 Not Modified')
        self.assertEqual(lines[1], b'Connection: close')
        self.assertTrue(lines[2].startswith(b'Date:'))
        self.assertEqual(lines[3], b'Server: waitress')
        self.assertEqual(inst.close_on_finish, True)
        self.assertTrue(('Connection', 'close') in inst.response_headers)

    def test_build_response_header_via_added(self):
        inst = self._makeOne()
        inst.request = DummyParser()
        inst.version = '1.0'
        inst.response_headers = [('Server', 'abc')]
        result = inst.build_response_header()
        lines = filter_lines(result)
        self.assertEqual(len(lines), 5)
        self.assertEqual(lines[0], b'HTTP/1.0 200 OK')
        self.assertEqual(lines[1], b'Connection: close')
        self.assertTrue(lines[2].startswith(b'Date:'))
        self.assertEqual(lines[3], b'Server: abc')
        self.assertEqual(lines[4], b'Via: waitress')

    def test_build_response_header_date_exists(self):
        inst = self._makeOne()
        inst.request = DummyParser()
        inst.version = '1.0'
        inst.response_headers = [('Date', 'date')]
        result = inst.build_response_header()
        lines = filter_lines(result)
        self.assertEqual(len(lines), 4)
        self.assertEqual(lines[0], b'HTTP/1.0 200 OK')
        self.assertEqual(lines[1], b'Connection: close')
        self.assertTrue(lines[2].startswith(b'Date:'))
        self.assertEqual(lines[3], b'Server: waitress')

    def test_build_response_header_preexisting_content_length(self):
        inst = self._makeOne()
        inst.request = DummyParser()
        inst.version = '1.1'
        inst.content_length = 100
        result = inst.build_response_header()
        lines = filter_lines(result)
        self.assertEqual(len(lines), 4)
        self.assertEqual(lines[0], b'HTTP/1.1 200 OK')
        self.assertEqual(lines[1], b'Content-Length: 100')
        self.assertTrue(lines[2].startswith(b'Date:'))
        self.assertEqual(lines[3], b'Server: waitress')

    def test_remove_content_length_header(self):
        inst = self._makeOne()
        inst.response_headers = [('Content-Length', '70')]
        inst.remove_content_length_header()
        self.assertEqual(inst.response_headers, [])

    def test_remove_content_length_header_with_other(self):
        inst = self._makeOne()
        inst.response_headers = [('Content-Length', '70'), ('Content-Type', 'text/html')]
        inst.remove_content_length_header()
        self.assertEqual(inst.response_headers, [('Content-Type', 'text/html')])

    def test_start(self):
        inst = self._makeOne()
        inst.start()
        self.assertTrue(inst.start_time)

    def test_finish_didnt_write_header(self):
        inst = self._makeOne()
        inst.wrote_header = False
        inst.complete = True
        inst.finish()
        self.assertTrue(inst.channel.written)

    def test_finish_wrote_header(self):
        inst = self._makeOne()
        inst.wrote_header = True
        inst.finish()
        self.assertFalse(inst.channel.written)

    def test_finish_chunked_response(self):
        inst = self._makeOne()
        inst.wrote_header = True
        inst.chunked_response = True
        inst.finish()
        self.assertEqual(inst.channel.written, b'0\r\n\r\n')

    def test_write_wrote_header(self):
        inst = self._makeOne()
        inst.wrote_header = True
        inst.complete = True
        inst.content_length = 3
        inst.write(b'abc')
        self.assertEqual(inst.channel.written, b'abc')

    def test_write_header_not_written(self):
        inst = self._makeOne()
        inst.wrote_header = False
        inst.complete = True
        inst.write(b'abc')
        self.assertTrue(inst.channel.written)
        self.assertEqual(inst.wrote_header, True)

    def test_write_start_response_uncalled(self):
        inst = self._makeOne()
        self.assertRaises(RuntimeError, inst.write, b'')

    def test_write_chunked_response(self):
        inst = self._makeOne()
        inst.wrote_header = True
        inst.chunked_response = True
        inst.complete = True
        inst.write(b'abc')
        self.assertEqual(inst.channel.written, b'3\r\nabc\r\n')

    def test_write_preexisting_content_length(self):
        inst = self._makeOne()
        inst.wrote_header = True
        inst.complete = True
        inst.content_length = 1
        inst.logger = DummyLogger()
        inst.write(b'abc')
        self.assertTrue(inst.channel.written)
        self.assertEqual(inst.logged_write_excess, True)
        self.assertEqual(len(inst.logger.logged), 1)

class TestWSGITask(unittest.TestCase):

    def _makeOne(self, channel=None, request=None):
        if channel is None:
            channel = DummyChannel()
        if request is None:
            request = DummyParser()
        from waitress.task import WSGITask
        return WSGITask(channel, request)

    def test_service(self):
        inst = self._makeOne()
        def execute():
            inst.executed = True
        inst.execute = execute
        inst.complete = True
        inst.service()
        self.assertTrue(inst.start_time)
        self.assertTrue(inst.close_on_finish)
        self.assertTrue(inst.channel.written)
        self.assertEqual(inst.executed, True)

    def test_service_server_raises_socket_error(self):
        import socket
        inst = self._makeOne()
        def execute():
            raise socket.error
        inst.execute = execute
        self.assertRaises(socket.error, inst.service)
        self.assertTrue(inst.start_time)
        self.assertTrue(inst.close_on_finish)
        self.assertFalse(inst.channel.written)

    def test_execute_app_calls_start_response_twice_wo_exc_info(self):
        def app(environ, start_response):
            start_response('200 OK', [])
            start_response('200 OK', [])
        inst = self._makeOne()
        inst.channel.server.application = app
        self.assertRaises(AssertionError, inst.execute)

    def test_execute_app_calls_start_response_w_exc_info_complete(self):
        def app(environ, start_response):
            start_response('200 OK', [], [ValueError, ValueError(), None])
            return [b'a']
        inst = self._makeOne()
        inst.complete = True
        inst.channel.server.application = app
        inst.execute()
        self.assertTrue(inst.complete)
        self.assertEqual(inst.status, '200 OK')
        self.assertTrue(inst.channel.written)

    def test_execute_app_calls_start_response_w_excinf_headers_unwritten(self):
        def app(environ, start_response):
            start_response('200 OK', [], [ValueError, None, None])
            return [b'a']
        inst = self._makeOne()
        inst.wrote_header = False
        inst.channel.server.application = app
        inst.response_headers = [('a', 'b')]
        inst.execute()
        self.assertTrue(inst.complete)
        self.assertEqual(inst.status, '200 OK')
        self.assertTrue(inst.channel.written)
        self.assertFalse(('a','b') in inst.response_headers)

    def test_execute_app_calls_start_response_w_excinf_headers_written(self):
        def app(environ, start_response):
            start_response('200 OK', [], [ValueError, ValueError(), None])
        inst = self._makeOne()
        inst.complete = True
        inst.wrote_header = True
        inst.channel.server.application = app
        self.assertRaises(ValueError, inst.execute)

    def test_execute_bad_header_key(self):
        def app(environ, start_response):
            start_response('200 OK', [(None, 'a')])
        inst = self._makeOne()
        inst.channel.server.application = app
        self.assertRaises(AssertionError, inst.execute)

    def test_execute_bad_header_value(self):
        def app(environ, start_response):
            start_response('200 OK', [('a', None)])
        inst = self._makeOne()
        inst.channel.server.application = app
        self.assertRaises(AssertionError, inst.execute)

    def test_execute_hopbyhop_header(self):
        def app(environ, start_response):
            start_response('200 OK', [('Connection', 'close')])
        inst = self._makeOne()
        inst.channel.server.application = app
        self.assertRaises(AssertionError, inst.execute)

    def test_execute_bad_header_value_control_characters(self):
        def app(environ, start_response):
            start_response('200 OK', [('a', '\n')])
        inst = self._makeOne()
        inst.channel.server.application = app
        self.assertRaises(ValueError, inst.execute)

    def test_execute_bad_header_name_control_characters(self):
        def app(environ, start_response):
            start_response('200 OK', [('a\r', 'value')])
        inst = self._makeOne()
        inst.channel.server.application = app
        self.assertRaises(ValueError, inst.execute)

    def test_execute_bad_status_control_characters(self):
        def app(environ, start_response):
            start_response('200 OK\r', [])
        inst = self._makeOne()
        inst.channel.server.application = app
        self.assertRaises(ValueError, inst.execute)

    def test_preserve_header_value_order(self):
        def app(environ, start_response):
            write = start_response('200 OK', [('C', 'b'), ('A', 'b'), ('A', 'a')])
            write(b'abc')
            return []
        inst = self._makeOne()
        inst.channel.server.application = app
        inst.execute()
        self.assertTrue(b'A: b\r\nA: a\r\nC: b\r\n' in inst.channel.written)

    def test_execute_bad_status_value(self):
        def app(environ, start_response):
            start_response(None, [])
        inst = self._makeOne()
        inst.channel.server.application = app
        self.assertRaises(AssertionError, inst.execute)

    def test_execute_with_content_length_header(self):
        def app(environ, start_response):
            start_response('200 OK', [('Content-Length', '1')])
            return [b'a']
        inst = self._makeOne()
        inst.channel.server.application = app
        inst.execute()
        self.assertEqual(inst.content_length, 1)

    def test_execute_app_calls_write(self):
        def app(environ, start_response):
            write = start_response('200 OK', [('Content-Length', '3')])
            write(b'abc')
            return []
        inst = self._makeOne()
        inst.channel.server.application = app
        inst.execute()
        self.assertEqual(inst.channel.written[-3:], b'abc')

    def test_execute_app_returns_len1_chunk_without_cl(self):
        def app(environ, start_response):
            start_response('200 OK', [])
            return [b'abc']
        inst = self._makeOne()
        inst.channel.server.application = app
        inst.execute()
        self.assertEqual(inst.content_length, 3)

    def test_execute_app_returns_empty_chunk_as_first(self):
        def app(environ, start_response):
            start_response('200 OK', [])
            return ['', b'abc']
        inst = self._makeOne()
        inst.channel.server.application = app
        inst.execute()
        self.assertEqual(inst.content_length, None)

    def test_execute_app_returns_too_many_bytes(self):
        def app(environ, start_response):
            start_response('200 OK', [('Content-Length', '1')])
            return [b'abc']
        inst = self._makeOne()
        inst.channel.server.application = app
        inst.logger = DummyLogger()
        inst.execute()
        self.assertEqual(inst.close_on_finish, True)
        self.assertEqual(len(inst.logger.logged), 1)

    def test_execute_app_returns_too_few_bytes(self):
        def app(environ, start_response):
            start_response('200 OK', [('Content-Length', '3')])
            return [b'a']
        inst = self._makeOne()
        inst.channel.server.application = app
        inst.logger = DummyLogger()
        inst.execute()
        self.assertEqual(inst.close_on_finish, True)
        self.assertEqual(len(inst.logger.logged), 1)

    def test_execute_app_do_not_warn_on_head(self):
        def app(environ, start_response):
            start_response('200 OK', [('Content-Length', '3')])
            return [b'']
        inst = self._makeOne()
        inst.request.command = 'HEAD'
        inst.channel.server.application = app
        inst.logger = DummyLogger()
        inst.execute()
        self.assertEqual(inst.close_on_finish, True)
        self.assertEqual(len(inst.logger.logged), 0)

    def test_execute_app_without_body_204_logged(self):
        def app(environ, start_response):
            start_response('204 No Content', [('Content-Length', '3')])
            return [b'abc']
        inst = self._makeOne()
        inst.channel.server.application = app
        inst.logger = DummyLogger()
        inst.execute()
        self.assertEqual(inst.close_on_finish, True)
        self.assertNotIn(b'abc', inst.channel.written)
        self.assertNotIn(b'Content-Length', inst.channel.written)
        self.assertNotIn(b'Transfer-Encoding', inst.channel.written)
        self.assertEqual(len(inst.logger.logged), 1)

    def test_execute_app_without_body_304_logged(self):
        def app(environ, start_response):
            start_response('304 Not Modified', [('Content-Length', '3')])
            return [b'abc']
        inst = self._makeOne()
        inst.channel.server.application = app
        inst.logger = DummyLogger()
        inst.execute()
        self.assertEqual(inst.close_on_finish, True)
        self.assertNotIn(b'abc', inst.channel.written)
        self.assertNotIn(b'Content-Length', inst.channel.written)
        self.assertNotIn(b'Transfer-Encoding', inst.channel.written)
        self.assertEqual(len(inst.logger.logged), 1)

    def test_execute_app_returns_closeable(self):
        class closeable(list):
            def close(self):
                self.closed = True
        foo = closeable([b'abc'])
        def app(environ, start_response):
            start_response('200 OK', [('Content-Length', '3')])
            return foo
        inst = self._makeOne()
        inst.channel.server.application = app
        inst.execute()
        self.assertEqual(foo.closed, True)

    def test_execute_app_returns_filewrapper_prepare_returns_True(self):
        from waitress.buffers import ReadOnlyFileBasedBuffer
        f = io.BytesIO(b'abc')
        app_iter = ReadOnlyFileBasedBuffer(f, 8192)
        def app(environ, start_response):
            start_response('200 OK', [('Content-Length', '3')])
            return app_iter
        inst = self._makeOne()
        inst.channel.server.application = app
        inst.execute()
        self.assertTrue(inst.channel.written) # header
        self.assertEqual(inst.channel.otherdata, [app_iter])

    def test_execute_app_returns_filewrapper_prepare_returns_True_nocl(self):
        from waitress.buffers import ReadOnlyFileBasedBuffer
        f = io.BytesIO(b'abc')
        app_iter = ReadOnlyFileBasedBuffer(f, 8192)
        def app(environ, start_response):
            start_response('200 OK', [])
            return app_iter
        inst = self._makeOne()
        inst.channel.server.application = app
        inst.execute()
        self.assertTrue(inst.channel.written) # header
        self.assertEqual(inst.channel.otherdata, [app_iter])
        self.assertEqual(inst.content_length, 3)

    def test_execute_app_returns_filewrapper_prepare_returns_True_badcl(self):
        from waitress.buffers import ReadOnlyFileBasedBuffer
        f = io.BytesIO(b'abc')
        app_iter = ReadOnlyFileBasedBuffer(f, 8192)
        def app(environ, start_response):
            start_response('200 OK', [])
            return app_iter
        inst = self._makeOne()
        inst.channel.server.application = app
        inst.content_length = 10
        inst.response_headers = [('Content-Length', '10')]
        inst.execute()
        self.assertTrue(inst.channel.written) # header
        self.assertEqual(inst.channel.otherdata, [app_iter])
        self.assertEqual(inst.content_length, 3)
        self.assertEqual(dict(inst.response_headers)['Content-Length'], '3')

    def test_get_environment_already_cached(self):
        inst = self._makeOne()
        inst.environ = object()
        self.assertEqual(inst.get_environment(), inst.environ)

    def test_get_environment_path_startswith_more_than_one_slash(self):
        inst = self._makeOne()
        request = DummyParser()
        request.path = '///abc'
        inst.request = request
        environ = inst.get_environment()
        self.assertEqual(environ['PATH_INFO'], '/abc')

    def test_get_environment_path_empty(self):
        inst = self._makeOne()
        request = DummyParser()
        request.path = ''
        inst.request = request
        environ = inst.get_environment()
        self.assertEqual(environ['PATH_INFO'], '')

    def test_get_environment_no_query(self):
        inst = self._makeOne()
        request = DummyParser()
        inst.request = request
        environ = inst.get_environment()
        self.assertEqual(environ['QUERY_STRING'], '')

    def test_get_environment_with_query(self):
        inst = self._makeOne()
        request = DummyParser()
        request.query = 'abc'
        inst.request = request
        environ = inst.get_environment()
        self.assertEqual(environ['QUERY_STRING'], 'abc')

    def test_get_environ_with_url_prefix_miss(self):
        inst = self._makeOne()
        inst.channel.server.adj.url_prefix = '/foo'
        request = DummyParser()
        request.path = '/bar'
        inst.request = request
        environ = inst.get_environment()
        self.assertEqual(environ['PATH_INFO'], '/bar')
        self.assertEqual(environ['SCRIPT_NAME'], '/foo')

    def test_get_environ_with_url_prefix_hit(self):
        inst = self._makeOne()
        inst.channel.server.adj.url_prefix = '/foo'
        request = DummyParser()
        request.path = '/foo/fuz'
        inst.request = request
        environ = inst.get_environment()
        self.assertEqual(environ['PATH_INFO'], '/fuz')
        self.assertEqual(environ['SCRIPT_NAME'], '/foo')

    def test_get_environ_with_url_prefix_empty_path(self):
        inst = self._makeOne()
        inst.channel.server.adj.url_prefix = '/foo'
        request = DummyParser()
        request.path = '/foo'
        inst.request = request
        environ = inst.get_environment()
        self.assertEqual(environ['PATH_INFO'], '')
        self.assertEqual(environ['SCRIPT_NAME'], '/foo')

    def test_get_environment_values(self):
        import sys
        inst = self._makeOne()
        request = DummyParser()
        request.headers = {
            'CONTENT_TYPE': 'abc',
            'CONTENT_LENGTH': '10',
            'X_FOO': 'BAR',
            'CONNECTION': 'close',
        }
        request.query = 'abc'
        inst.request = request
        environ = inst.get_environment()

        # nail the keys of environ
        self.assertEqual(sorted(environ.keys()), [
            'CONTENT_LENGTH', 'CONTENT_TYPE', 'HTTP_CONNECTION', 'HTTP_X_FOO',
            'PATH_INFO', 'QUERY_STRING', 'REMOTE_ADDR', 'REMOTE_HOST',
            'REMOTE_PORT', 'REQUEST_METHOD', 'SCRIPT_NAME', 'SERVER_NAME',
            'SERVER_PORT', 'SERVER_PROTOCOL', 'SERVER_SOFTWARE', 'wsgi.errors',
            'wsgi.file_wrapper', 'wsgi.input', 'wsgi.input_terminated',
            'wsgi.multiprocess', 'wsgi.multithread', 'wsgi.run_once',
            'wsgi.url_scheme', 'wsgi.version'
        ])

        self.assertEqual(environ['REQUEST_METHOD'], 'GET')
        self.assertEqual(environ['SERVER_PORT'], '80')
        self.assertEqual(environ['SERVER_NAME'], 'localhost')
        self.assertEqual(environ['SERVER_SOFTWARE'], 'waitress')
        self.assertEqual(environ['SERVER_PROTOCOL'], 'HTTP/1.0')
        self.assertEqual(environ['SCRIPT_NAME'], '')
        self.assertEqual(environ['HTTP_CONNECTION'], 'close')
        self.assertEqual(environ['PATH_INFO'], '/')
        self.assertEqual(environ['QUERY_STRING'], 'abc')
        self.assertEqual(environ['REMOTE_ADDR'], '127.0.0.1')
        self.assertEqual(environ['REMOTE_HOST'], '127.0.0.1')
        self.assertEqual(environ['REMOTE_PORT'], '39830')
        self.assertEqual(environ['CONTENT_TYPE'], 'abc')
        self.assertEqual(environ['CONTENT_LENGTH'], '10')
        self.assertEqual(environ['HTTP_X_FOO'], 'BAR')
        self.assertEqual(environ['wsgi.version'], (1, 0))
        self.assertEqual(environ['wsgi.url_scheme'], 'http')
        self.assertEqual(environ['wsgi.errors'], sys.stderr)
        self.assertEqual(environ['wsgi.multithread'], True)
        self.assertEqual(environ['wsgi.multiprocess'], False)
        self.assertEqual(environ['wsgi.run_once'], False)
        self.assertEqual(environ['wsgi.input'], 'stream')
        self.assertEqual(environ['wsgi.input_terminated'], True)
        self.assertEqual(inst.environ, environ)

    def test_get_environment_values_w_scheme_override_untrusted(self):
        inst = self._makeOne()
        request = DummyParser()
        request.headers = {
            'CONTENT_TYPE': 'abc',
            'CONTENT_LENGTH': '10',
            'X_FOO': 'BAR',
            'X_FORWARDED_PROTO': 'https',
            'CONNECTION': 'close',
        }
        request.query = 'abc'
        inst.request = request
        environ = inst.get_environment()
        self.assertEqual(environ['wsgi.url_scheme'], 'http')

    def test_get_environment_values_w_scheme_override_trusted(self):
        import sys
        inst = self._makeOne()
        inst.channel.addr = ['192.168.1.1', 8080]
        inst.channel.server.adj.trusted_proxy = '192.168.1.1'
        inst.channel.server.adj.trusted_proxy_headers = {'x-forwarded-proto'}
        request = DummyParser()
        request.headers = {
            'CONTENT_TYPE': 'abc',
            'CONTENT_LENGTH': '10',
            'X_FOO': 'BAR',
            'X_FORWARDED_PROTO': 'https',
            'CONNECTION': 'close',
        }
        request.query = 'abc'
        inst.request = request
        environ = inst.get_environment()

        # nail the keys of environ
        self.assertEqual(sorted(environ.keys()), [
            'CONTENT_LENGTH', 'CONTENT_TYPE', 'HTTP_CONNECTION', 'HTTP_X_FOO',
            'HTTP_X_FORWARDED_PROTO', 'PATH_INFO', 'QUERY_STRING',
            'REMOTE_ADDR', 'REMOTE_HOST', 'REQUEST_METHOD', 'SCRIPT_NAME',
            'SERVER_NAME', 'SERVER_PORT', 'SERVER_PROTOCOL', 'SERVER_SOFTWARE',
            'wsgi.errors', 'wsgi.file_wrapper', 'wsgi.input',
            'wsgi.input_terminated', 'wsgi.multiprocess', 'wsgi.multithread',
            'wsgi.run_once', 'wsgi.url_scheme', 'wsgi.version'
            ])

        self.assertEqual(environ['REQUEST_METHOD'], 'GET')
        self.assertEqual(environ['SERVER_PORT'], '443')
        self.assertEqual(environ['SERVER_NAME'], 'localhost')
        self.assertEqual(environ['SERVER_SOFTWARE'], 'waitress')
        self.assertEqual(environ['SERVER_PROTOCOL'], 'HTTP/1.0')
        self.assertEqual(environ['SCRIPT_NAME'], '')
        self.assertEqual(environ['HTTP_CONNECTION'], 'close')
        self.assertEqual(environ['PATH_INFO'], '/')
        self.assertEqual(environ['QUERY_STRING'], 'abc')
        self.assertEqual(environ['REMOTE_ADDR'], '192.168.1.1')
        self.assertEqual(environ['CONTENT_TYPE'], 'abc')
        self.assertEqual(environ['CONTENT_LENGTH'], '10')
        self.assertEqual(environ['HTTP_X_FOO'], 'BAR')
        self.assertEqual(environ['wsgi.version'], (1, 0))
        self.assertEqual(environ['wsgi.url_scheme'], 'https')
        self.assertEqual(environ['wsgi.errors'], sys.stderr)
        self.assertEqual(environ['wsgi.multithread'], True)
        self.assertEqual(environ['wsgi.multiprocess'], False)
        self.assertEqual(environ['wsgi.run_once'], False)
        self.assertEqual(environ['wsgi.input'], 'stream')
        self.assertEqual(environ['wsgi.input_terminated'], True)
        self.assertEqual(inst.environ, environ)

    def test_get_environment_values_w_bogus_scheme_override(self):
        inst = self._makeOne()
        inst.channel.addr = ['192.168.1.1', 80]
        inst.channel.server.adj.trusted_proxy = '192.168.1.1'
        inst.channel.server.adj.trusted_proxy_headers = {'x-forwarded-proto'}
        request = DummyParser()
        request.headers = {
            'CONTENT_TYPE': 'abc',
            'CONTENT_LENGTH': '10',
            'X_FOO': 'BAR',
            'X_FORWARDED_PROTO': 'http://p02n3e.com?url=http',
            'CONNECTION': 'close',
        }
        request.query = 'abc'
        inst.request = request
        self.assertRaises(ValueError, inst.get_environment)

    def test_get_environment_warning_other_proxy_headers(self):
        inst = self._makeOne()
        inst.logger = DummyLogger()

        inst.request.headers = {
            'X_FORWARDED_FOR': '[2001:db8::1]',
            'FORWARDED': 'For=198.51.100.2;host=example.com:8080;proto=https'
        }
        inst.channel.addr = ['192.168.1.1', 80]
        inst.channel.server.adj.trusted_proxy = '192.168.1.1'
        inst.channel.server.adj.trusted_proxy_count = 1
        inst.channel.server.adj.trusted_proxy_headers = {'forwarded'}
        inst.channel.server.adj.log_untrusted_proxy_headers = True
        environ = inst.get_environment()

        self.assertEqual(len(inst.logger.logged), 1)
        self.assertNotIn('HTTP_X_FORWARDED_FOR', environ)

        self.assertEqual(environ['REMOTE_ADDR'], '198.51.100.2')
        self.assertEqual(environ['SERVER_NAME'], 'example.com')
        self.assertEqual(environ['HTTP_HOST'], 'example.com:8080')
        self.assertEqual(environ['SERVER_PORT'], '8080')
        self.assertEqual(environ['wsgi.url_scheme'], 'https')

    def test_get_environment_contains_all_headers_including_untrusted(self):
        inst = self._makeOne()
        inst.logger = DummyLogger()

        inst.request.headers = {
            'X_FORWARDED_FOR': '198.51.100.2',
            'X_FORWARDED_BY': 'Waitress',
            'X_FORWARDED_PROTO': 'https',
            'X_FORWARDED_HOST': 'example.org',
        }
        headers_orig = inst.request.headers.copy()
        inst.channel.addr = ['192.168.1.1', 80]
        inst.channel.server.adj.trusted_proxy = '192.168.1.1'
        inst.channel.server.adj.trusted_proxy_count = 1
        inst.channel.server.adj.trusted_proxy_headers = {'x-forwarded-by'}
        inst.channel.server.adj.clear_untrusted_proxy_headers = False
        environ = inst.get_environment()

        for k, expected in headers_orig.items():
            result = environ['HTTP_%s' % k]
            self.assertEqual(result, expected)

    def test_get_environment_contains_only_trusted_headers(self):
        inst = self._makeOne()
        inst.logger = DummyLogger()

        inst.request.headers = {
            'X_FORWARDED_FOR': '198.51.100.2',
            'X_FORWARDED_BY': 'Waitress',
            'X_FORWARDED_PROTO': 'https',
            'X_FORWARDED_HOST': 'example.org',
        }
        inst.channel.addr = ['192.168.1.1', 80]
        inst.channel.server.adj.trusted_proxy = '192.168.1.1'
        inst.channel.server.adj.trusted_proxy_count = 1
        inst.channel.server.adj.trusted_proxy_headers = {'x-forwarded-by'}
        inst.channel.server.adj.clear_untrusted_proxy_headers = True
        environ = inst.get_environment()

        self.assertEqual(environ['HTTP_X_FORWARDED_BY'], 'Waitress')
        self.assertNotIn('HTTP_X_FORWARDED_FOR', environ)
        self.assertNotIn('HTTP_X_FORWARDED_PROTO', environ)
        self.assertNotIn('HTTP_X_FORWARDED_HOST', environ)

    def test_get_environment_clears_headers_if_untrusted_proxy(self):
        inst = self._makeOne()
        inst.logger = DummyLogger()

        inst.request.headers = {
            'X_FORWARDED_FOR': '198.51.100.2',
            'X_FORWARDED_BY': 'Waitress',
            'X_FORWARDED_PROTO': 'https',
            'X_FORWARDED_HOST': 'example.org',
        }
        inst.channel.addr = ['192.168.1.255', 80]
        inst.channel.server.adj.trusted_proxy = '192.168.1.1'
        inst.channel.server.adj.trusted_proxy_count = 1
        inst.channel.server.adj.trusted_proxy_headers = {'x-forwarded-by'}
        inst.channel.server.adj.clear_untrusted_proxy_headers = True
        environ = inst.get_environment()

        self.assertNotIn('HTTP_X_FORWARDED_BY', environ)
        self.assertNotIn('HTTP_X_FORWARDED_FOR', environ)
        self.assertNotIn('HTTP_X_FORWARDED_PROTO', environ)
        self.assertNotIn('HTTP_X_FORWARDED_HOST', environ)

    def test_parse_proxy_headers_forwarded_for(self):
        inst = self._makeOne()

        headers = {
            'X_FORWARDED_FOR': '192.0.2.1'
        }
        environ = {}
        inst.parse_proxy_headers(
            environ,
            headers,
            trusted_proxy_count=1,
            trusted_proxy_headers={'x-forwarded-for'}
        )

        self.assertEqual(environ['REMOTE_ADDR'], '192.0.2.1')

    def test_parse_proxy_headers_forwarded_for_v6_missing_brackets(self):
        inst = self._makeOne()

        headers = {
            'X_FORWARDED_FOR': '2001:db8::0'
        }
        environ = {}
        inst.parse_proxy_headers(
            environ,
            headers,
            trusted_proxy_count=1,
            trusted_proxy_headers={'x-forwarded-for'}
        )

        self.assertEqual(environ['REMOTE_ADDR'], '2001:db8::0')

    def test_parse_proxy_headers_forwared_for_multiple(self):
        inst = self._makeOne()

        headers = {
            'X_FORWARDED_FOR': '192.0.2.1, 198.51.100.2, 203.0.113.1'
        }
        environ = {}
        inst.parse_proxy_headers(
            environ,
            headers,
            trusted_proxy_count=2,
            trusted_proxy_headers={'x-forwarded-for'}
        )
        self.assertEqual(environ['REMOTE_ADDR'], '198.51.100.2')

    def test_parse_forwarded_multiple_proxies_trust_only_two(self):
        inst = self._makeOne()

        headers = {
            'FORWARDED': 'For=192.0.2.1;host=fake.com, For=198.51.100.2;host=example.com:8080, For=203.0.113.1'
        }
        environ = {}
        inst.parse_proxy_headers(
            environ,
            headers,
            trusted_proxy_count=2,
            trusted_proxy_headers={'forwarded'}
        )

        self.assertEqual(environ['REMOTE_ADDR'], '198.51.100.2')
        self.assertEqual(environ['SERVER_NAME'], 'example.com')
        self.assertEqual(environ['HTTP_HOST'], 'example.com:8080')
        self.assertEqual(environ['SERVER_PORT'], '8080')

    def test_parse_forwarded_multiple_proxies(self):
        inst = self._makeOne()

        headers = {
            'FORWARDED': 'for="[2001:db8::1]:3821";host="example.com:8443";proto="https", for=192.0.2.1;host="example.internal:8080"'
        }
        environ = {}
        inst.parse_proxy_headers(
            environ,
            headers,
            trusted_proxy_count=2,
            trusted_proxy_headers={'forwarded'}
        )

        self.assertEqual(environ['REMOTE_ADDR'], '2001:db8::1')
        self.assertEqual(environ['REMOTE_PORT'], '3821')
        self.assertEqual(environ['SERVER_NAME'], 'example.com')
        self.assertEqual(environ['HTTP_HOST'], 'example.com:8443')
        self.assertEqual(environ['SERVER_PORT'], '8443')
        self.assertEqual(environ['wsgi.url_scheme'], 'https')

    def test_parse_forwarded_multiple_proxies_minimal(self):
        inst = self._makeOne()

        headers = {
            'FORWARDED': 'for="[2001:db8::1]";proto="https", for=192.0.2.1;host="example.org"'
        }
        environ = {}
        inst.parse_proxy_headers(
            environ,
            headers,
            trusted_proxy_count=2,
            trusted_proxy_headers={'forwarded'}
        )

        self.assertEqual(environ['REMOTE_ADDR'], '2001:db8::1')
        self.assertEqual(environ['SERVER_NAME'], 'example.org')
        self.assertEqual(environ['HTTP_HOST'], 'example.org')
        self.assertEqual(environ['SERVER_PORT'], '443')
        self.assertEqual(environ['wsgi.url_scheme'], 'https')

    def test_parse_proxy_headers_forwarded_host_with_port(self):
        inst = self._makeOne()

        headers = {
            'X_FORWARDED_FOR': '192.0.2.1, 198.51.100.2, 203.0.113.1',
            'X_FORWARDED_PROTO': 'http',
            'X_FORWARDED_HOST': 'example.com:8080',
        }
        environ = {}
        inst.parse_proxy_headers(
            environ,
            headers,
            trusted_proxy_count=2,
            trusted_proxy_headers={'x-forwarded-for', 'x-forwarded-proto', 'x-forwarded-host'}
        )

        self.assertEqual(environ['REMOTE_ADDR'], '198.51.100.2')
        self.assertEqual(environ['SERVER_NAME'], 'example.com')
        self.assertEqual(environ['HTTP_HOST'], 'example.com:8080')
        self.assertEqual(environ['SERVER_PORT'], '8080')

    def test_parse_proxy_headers_forwarded_host_without_port(self):
        inst = self._makeOne()

        headers = {
            'X_FORWARDED_FOR': '192.0.2.1, 198.51.100.2, 203.0.113.1',
            'X_FORWARDED_PROTO': 'http',
            'X_FORWARDED_HOST': 'example.com',
        }
        environ = {}
        inst.parse_proxy_headers(
            environ,
            headers,
            trusted_proxy_count=2,
            trusted_proxy_headers={'x-forwarded-for', 'x-forwarded-proto', 'x-forwarded-host'}
        )

        self.assertEqual(environ['REMOTE_ADDR'], '198.51.100.2')
        self.assertEqual(environ['SERVER_NAME'], 'example.com')
        self.assertEqual(environ['HTTP_HOST'], 'example.com')
        self.assertEqual(environ['SERVER_PORT'], '80')

    def test_parse_proxy_headers_forwarded_host_with_forwarded_port(self):
        inst = self._makeOne()

        headers = {
            'X_FORWARDED_FOR': '192.0.2.1, 198.51.100.2, 203.0.113.1',
            'X_FORWARDED_PROTO': 'http',
            'X_FORWARDED_HOST': 'example.com',
            'X_FORWARDED_PORT': '8080'
        }
        environ = {}
        inst.parse_proxy_headers(
            environ,
            headers,
            trusted_proxy_count=2,
            trusted_proxy_headers={'x-forwarded-for', 'x-forwarded-proto', 'x-forwarded-host', 'x-forwarded-port'}
        )

        self.assertEqual(environ['REMOTE_ADDR'], '198.51.100.2')
        self.assertEqual(environ['SERVER_NAME'], 'example.com')
        self.assertEqual(environ['HTTP_HOST'], 'example.com:8080')
        self.assertEqual(environ['SERVER_PORT'], '8080')

    def test_parse_proxy_headers_forwarded_host_multiple_with_forwarded_port(self):
        inst = self._makeOne()

        headers = {
            'X_FORWARDED_FOR': '192.0.2.1, 198.51.100.2, 203.0.113.1',
            'X_FORWARDED_PROTO': 'http',
            'X_FORWARDED_HOST': 'example.com, example.org',
            'X_FORWARDED_PORT': '8080'
        }
        environ = {}
        inst.parse_proxy_headers(
            environ,
            headers,
            trusted_proxy_count=2,
            trusted_proxy_headers={'x-forwarded-for', 'x-forwarded-proto', 'x-forwarded-host', 'x-forwarded-port'}
        )

        self.assertEqual(environ['REMOTE_ADDR'], '198.51.100.2')
        self.assertEqual(environ['SERVER_NAME'], 'example.com')
        self.assertEqual(environ['HTTP_HOST'], 'example.com:8080')
        self.assertEqual(environ['SERVER_PORT'], '8080')

    def test_parse_proxy_headers_forwarded_host_multiple_with_forwarded_port_limit_one_trusted(self):
        inst = self._makeOne()

        headers = {
            'X_FORWARDED_FOR': '192.0.2.1, 198.51.100.2, 203.0.113.1',
            'X_FORWARDED_PROTO': 'http',
            'X_FORWARDED_HOST': 'example.com, example.org',
            'X_FORWARDED_PORT': '8080'
        }
        environ = {}
        inst.parse_proxy_headers(
            environ,
            headers,
            trusted_proxy_count=1,
            trusted_proxy_headers={'x-forwarded-for', 'x-forwarded-proto', 'x-forwarded-host', 'x-forwarded-port'}
        )

        self.assertEqual(environ['REMOTE_ADDR'], '203.0.113.1')
        self.assertEqual(environ['SERVER_NAME'], 'example.org')
        self.assertEqual(environ['HTTP_HOST'], 'example.org:8080')
        self.assertEqual(environ['SERVER_PORT'], '8080')

    def test_parse_forwarded(self):
        inst = self._makeOne()

        headers = {
            'FORWARDED': 'For=198.51.100.2:5858;host=example.com:8080;proto=https'
        }
        environ = {}
        inst.parse_proxy_headers(
            environ,
            headers,
            trusted_proxy_count=1,
            trusted_proxy_headers={'forwarded'}
        )

        self.assertEqual(environ['REMOTE_ADDR'], '198.51.100.2')
        self.assertEqual(environ['REMOTE_PORT'], '5858')
        self.assertEqual(environ['SERVER_NAME'], 'example.com')
        self.assertEqual(environ['HTTP_HOST'], 'example.com:8080')
        self.assertEqual(environ['SERVER_PORT'], '8080')
        self.assertEqual(environ['wsgi.url_scheme'], 'https')

    def test_parse_forwarded_empty_pair(self):
        inst = self._makeOne()

        headers = {
            'FORWARDED': 'For=198.51.100.2;;proto=https;by=_unused'
        }
        environ = {}
        inst.parse_proxy_headers(
            environ,
            headers,
            trusted_proxy_count=1,
            trusted_proxy_headers={'forwarded'}
        )

        self.assertEqual(environ['REMOTE_ADDR'], '198.51.100.2')

    def test_parse_forwarded_pair_token_whitespace(self):
        inst = self._makeOne()

        headers = {
            'FORWARDED': 'For=198.51.100.2; proto =https'
        }
        environ = {}

        with self.assertRaises(ValueError):
            inst.parse_proxy_headers(
                environ,
                headers,
                trusted_proxy_count=1,
                trusted_proxy_headers={'forwarded'}
            )

    def test_parse_forwarded_pair_value_whitespace(self):
        inst = self._makeOne()

        headers = {
            'FORWARDED': 'For= "198.51.100.2"; proto =https'
        }
        environ = {}

        with self.assertRaises(ValueError):
            inst.parse_proxy_headers(
                environ,
                headers,
                trusted_proxy_count=1,
                trusted_proxy_headers={'forwarded'}
            )

    def test_parse_forwarded_pair_no_equals(self):
        inst = self._makeOne()

        headers = {
            'FORWARDED': 'For'
        }
        environ = {}

        with self.assertRaises(ValueError):
            inst.parse_proxy_headers(
                environ,
                headers,
                trusted_proxy_count=1,
                trusted_proxy_headers={'forwarded'}
            )

    def test_parse_forwarded_warning_unknown_token(self):
        inst = self._makeOne()
        inst.logger = DummyLogger()

        headers = {
            'FORWARDED': 'For=198.51.100.2;host=example.com:8080;proto=https;unknown="yolo"'
        }
        environ = {}
        inst.parse_proxy_headers(
            environ,
            headers,
            trusted_proxy_count=1,
            trusted_proxy_headers={'forwarded'}
        )

        self.assertEqual(len(inst.logger.logged), 1)
        self.assertIn('Unknown Forwarded token', inst.logger.logged[0])

        self.assertEqual(environ['REMOTE_ADDR'], '198.51.100.2')
        self.assertEqual(environ['SERVER_NAME'], 'example.com')
        self.assertEqual(environ['HTTP_HOST'], 'example.com:8080')
        self.assertEqual(environ['SERVER_PORT'], '8080')
        self.assertEqual(environ['wsgi.url_scheme'], 'https')

    def test_parse_no_valid_proxy_headers(self):
        inst = self._makeOne()
        inst.logger = DummyLogger()

        headers = {
            'X_FORWARDED_FOR': '198.51.100.2',
            'FORWARDED': 'For=198.51.100.2;host=example.com:8080;proto=https'
        }
        environ = {}
        inst.parse_proxy_headers(
            environ,
            headers,
            trusted_proxy_count=1,
        )

        self.assertEqual(environ, {})

    def test_parse_multiple_x_forwarded_proto(self):
        inst = self._makeOne()
        inst.logger = DummyLogger()

        headers = {
            'X_FORWARDED_PROTO': 'http, https',
        }
        environ = {}
        inst.parse_proxy_headers(
            environ,
            headers,
            trusted_proxy_count=1,
            trusted_proxy_headers={'x-forwarded-proto'}
        )

        self.assertEqual(environ, {})
        self.assertEqual(len(inst.logger.logged), 1)
        self.assertIn("Found multiple values in X-Forwarded-Proto", inst.logger.logged[0])

    def test_parse_multiple_x_forwarded_port(self):
        inst = self._makeOne()
        inst.logger = DummyLogger()

        headers = {
            'X_FORWARDED_PORT': '443, 80',
        }
        environ = {}
        inst.parse_proxy_headers(
            environ,
            headers,
            trusted_proxy_count=1,
            trusted_proxy_headers={'x-forwarded-port'}
        )

        self.assertEqual(environ, {})
        self.assertEqual(len(inst.logger.logged), 1)
        self.assertIn("Found multiple values in X-Forwarded-Port", inst.logger.logged[0])

    def test_parse_forwarded_port_wrong_proto_port_80(self):
        inst = self._makeOne()
        inst.logger = DummyLogger()

        headers = {
            'X_FORWARDED_PORT': '80',
            'X_FORWARDED_PROTO': 'https',
            'X_FORWARDED_HOST': 'example.com',
        }
        environ = {}
        inst.parse_proxy_headers(
            environ,
            headers,
            trusted_proxy_count=1,
            trusted_proxy_headers={'x-forwarded-port', 'x-forwarded-host', 'x-forwarded-proto'}
        )

        self.assertEqual(environ['SERVER_NAME'], 'example.com')
        self.assertEqual(environ['HTTP_HOST'], 'example.com:80')
        self.assertEqual(environ['SERVER_PORT'], '80')
        self.assertEqual(environ['wsgi.url_scheme'], 'https')

    def test_parse_forwarded_port_wrong_proto_port_443(self):
        inst = self._makeOne()
        inst.logger = DummyLogger()

        headers = {
            'X_FORWARDED_PORT': '443',
            'X_FORWARDED_PROTO': 'http',
            'X_FORWARDED_HOST': 'example.com',
        }
        environ = {}
        inst.parse_proxy_headers(
            environ,
            headers,
            trusted_proxy_count=1,
            trusted_proxy_headers={'x-forwarded-port', 'x-forwarded-host', 'x-forwarded-proto'}
        )

        self.assertEqual(environ['SERVER_NAME'], 'example.com')
        self.assertEqual(environ['HTTP_HOST'], 'example.com:443')
        self.assertEqual(environ['SERVER_PORT'], '443')
        self.assertEqual(environ['wsgi.url_scheme'], 'http')


class TestErrorTask(unittest.TestCase):

    def _makeOne(self, channel=None, request=None):
        if channel is None:
            channel = DummyChannel()
        if request is None:
            request = DummyParser()
            request.error = DummyError()
        from waitress.task import ErrorTask
        return ErrorTask(channel, request)

    def test_execute_http_10(self):
        inst = self._makeOne()
        inst.execute()
        lines = filter_lines(inst.channel.written)
        self.assertEqual(len(lines), 9)
        self.assertEqual(lines[0], b'HTTP/1.0 432 Too Ugly')
        self.assertEqual(lines[1], b'Connection: close')
        self.assertEqual(lines[2], b'Content-Length: 43')
        self.assertEqual(lines[3], b'Content-Type: text/plain')
        self.assertTrue(lines[4])
        self.assertEqual(lines[5], b'Server: waitress')
        self.assertEqual(lines[6], b'Too Ugly')
        self.assertEqual(lines[7], b'body')
        self.assertEqual(lines[8], b'(generated by waitress)')

    def test_execute_http_11(self):
        inst = self._makeOne()
        inst.version = '1.1'
        inst.execute()
        lines = filter_lines(inst.channel.written)
        self.assertEqual(len(lines), 8)
        self.assertEqual(lines[0], b'HTTP/1.1 432 Too Ugly')
        self.assertEqual(lines[1], b'Content-Length: 43')
        self.assertEqual(lines[2], b'Content-Type: text/plain')
        self.assertTrue(lines[3])
        self.assertEqual(lines[4], b'Server: waitress')
        self.assertEqual(lines[5], b'Too Ugly')
        self.assertEqual(lines[6], b'body')
        self.assertEqual(lines[7], b'(generated by waitress)')

    def test_execute_http_11_close(self):
        inst = self._makeOne()
        inst.version = '1.1'
        inst.request.headers['CONNECTION'] = 'close'
        inst.execute()
        lines = filter_lines(inst.channel.written)
        self.assertEqual(len(lines), 9)
        self.assertEqual(lines[0], b'HTTP/1.1 432 Too Ugly')
        self.assertEqual(lines[1], b'Connection: close')
        self.assertEqual(lines[2], b'Content-Length: 43')
        self.assertEqual(lines[3], b'Content-Type: text/plain')
        self.assertTrue(lines[4])
        self.assertEqual(lines[5], b'Server: waitress')
        self.assertEqual(lines[6], b'Too Ugly')
        self.assertEqual(lines[7], b'body')
        self.assertEqual(lines[8], b'(generated by waitress)')

    def test_execute_http_11_keep(self):
        inst = self._makeOne()
        inst.version = '1.1'
        inst.request.headers['CONNECTION'] = 'keep-alive'
        inst.execute()
        lines = filter_lines(inst.channel.written)
        self.assertEqual(len(lines), 8)
        self.assertEqual(lines[0], b'HTTP/1.1 432 Too Ugly')
        self.assertEqual(lines[1], b'Content-Length: 43')
        self.assertEqual(lines[2], b'Content-Type: text/plain')
        self.assertTrue(lines[3])
        self.assertEqual(lines[4], b'Server: waitress')
        self.assertEqual(lines[5], b'Too Ugly')
        self.assertEqual(lines[6], b'body')
        self.assertEqual(lines[7], b'(generated by waitress)')


class DummyError(object):
    code = '432'
    reason = 'Too Ugly'
    body = 'body'

class DummyTask(object):
    serviced = False
    deferred = False
    cancelled = False

    def __init__(self, toraise=None):
        self.toraise = toraise

    def service(self):
        self.serviced = True
        if self.toraise:
            raise self.toraise

    def defer(self):
        self.deferred = True
        if self.toraise:
            raise self.toraise

    def cancel(self):
        self.cancelled = True

class DummyAdj(object):
    log_socket_errors = True
    ident = 'waitress'
    host = '127.0.0.1'
    port = 80
    url_prefix = ''
    trusted_proxy = None
    trusted_proxy_count = 1
    trusted_proxy_headers = set()
    log_untrusted_proxy_headers = True
    clear_untrusted_proxy_headers = True

class DummyServer(object):
    server_name = 'localhost'
    effective_port = 80

    def __init__(self):
        self.adj = DummyAdj()

class DummyChannel(object):
    closed_when_done = False
    adj = DummyAdj()
    creation_time = 0
    addr = ('127.0.0.1', 39830)

    def __init__(self, server=None):
        if server is None:
            server = DummyServer()
        self.server = server
        self.written = b''
        self.otherdata = []

    def write_soon(self, data):
        if isinstance(data, bytes):
            self.written += data
        else:
            self.otherdata.append(data)
        return len(data)

class DummyParser(object):
    version = '1.0'
    command = 'GET'
    path = '/'
    query = ''
    url_scheme = 'http'
    expect_continue = False
    headers_finished = False

    def __init__(self):
        self.headers = {}

    def get_body_stream(self):
        return 'stream'

def filter_lines(s):
    return list(filter(None, s.split(b'\r\n')))

class DummyLogger(object):

    def __init__(self):
        self.logged = []

    def warning(self, msg, *args):
        self.logged.append(msg % args)

    def exception(self, msg):
        self.logged.append(msg)



import sys
import socket
import warnings

from waitress.compat import (
    PY2,
    WIN,
    )

if sys.version_info[:2] == (2, 6): # pragma: no cover
    import unittest2 as unittest
else: # pragma: no cover
    import unittest

class Test_asbool(unittest.TestCase):

    def _callFUT(self, s):
        from waitress.adjustments import asbool
        return asbool(s)

    def test_s_is_None(self):
        result = self._callFUT(None)
        self.assertEqual(result, False)

    def test_s_is_True(self):
        result = self._callFUT(True)
        self.assertEqual(result, True)

    def test_s_is_False(self):
        result = self._callFUT(False)
        self.assertEqual(result, False)

    def test_s_is_true(self):
        result = self._callFUT('True')
        self.assertEqual(result, True)

    def test_s_is_false(self):
        result = self._callFUT('False')
        self.assertEqual(result, False)

    def test_s_is_yes(self):
        result = self._callFUT('yes')
        self.assertEqual(result, True)

    def test_s_is_on(self):
        result = self._callFUT('on')
        self.assertEqual(result, True)

    def test_s_is_1(self):
        result = self._callFUT(1)
        self.assertEqual(result, True)

class Test_as_socket_list(unittest.TestCase):

    def test_only_sockets_in_list(self):
        from waitress.adjustments import as_socket_list
        sockets = [
            socket.socket(socket.AF_INET, socket.SOCK_STREAM),
            socket.socket(socket.AF_INET6, socket.SOCK_STREAM)]
        if hasattr(socket, 'AF_UNIX'):
            sockets.append(socket.socket(socket.AF_UNIX, socket.SOCK_STREAM))
        new_sockets = as_socket_list(sockets)
        self.assertEqual(sockets, new_sockets)
        for sock in sockets:
            sock.close()

    def test_not_only_sockets_in_list(self):
        from waitress.adjustments import as_socket_list
        sockets = [
            socket.socket(socket.AF_INET, socket.SOCK_STREAM),
            socket.socket(socket.AF_INET6, socket.SOCK_STREAM),
            {'something': 'else'}]
        new_sockets = as_socket_list(sockets)
        self.assertEqual(new_sockets, [sockets[0], sockets[1]])
        for sock in [sock for sock in sockets if isinstance(sock, socket.socket)]:
            sock.close()

class TestAdjustments(unittest.TestCase):

    def _hasIPv6(self): # pragma: nocover
        if not socket.has_ipv6:
            return False

        try:
            socket.getaddrinfo(
                '::1',
                0,
                socket.AF_UNSPEC,
                socket.SOCK_STREAM,
                socket.IPPROTO_TCP,
                socket.AI_PASSIVE | socket.AI_ADDRCONFIG
            )

            return True
        except socket.gaierror as e:
            # Check to see what the error is
            if e.errno == socket.EAI_ADDRFAMILY:
                return False
            else:
                raise e

    def _makeOne(self, **kw):
        from waitress.adjustments import Adjustments
        return Adjustments(**kw)

    def test_goodvars(self):
        inst = self._makeOne(
            host='localhost',
            port='8080',
            threads='5',
            trusted_proxy='192.168.1.1',
            trusted_proxy_headers={'forwarded'},
            trusted_proxy_count=2,
            log_untrusted_proxy_headers=True,
            url_scheme='https',
            backlog='20',
            recv_bytes='200',
            send_bytes='300',
            outbuf_overflow='400',
            inbuf_overflow='500',
            connection_limit='1000',
            cleanup_interval='1100',
            channel_timeout='1200',
            log_socket_errors='true',
            max_request_header_size='1300',
            max_request_body_size='1400',
            expose_tracebacks='true',
            ident='abc',
            asyncore_loop_timeout='5',
            asyncore_use_poll=True,
            unix_socket_perms='777',
            url_prefix='///foo/',
            ipv4=True,
            ipv6=False,
        )

        self.assertEqual(inst.host, 'localhost')
        self.assertEqual(inst.port, 8080)
        self.assertEqual(inst.threads, 5)
        self.assertEqual(inst.trusted_proxy, '192.168.1.1')
        self.assertEqual(inst.trusted_proxy_headers, {'forwarded'})
        self.assertEqual(inst.trusted_proxy_count, 2)
        self.assertEqual(inst.log_untrusted_proxy_headers, True)
        self.assertEqual(inst.url_scheme, 'https')
        self.assertEqual(inst.backlog, 20)
        self.assertEqual(inst.recv_bytes, 200)
        self.assertEqual(inst.send_bytes, 300)
        self.assertEqual(inst.outbuf_overflow, 400)
        self.assertEqual(inst.inbuf_overflow, 500)
        self.assertEqual(inst.connection_limit, 1000)
        self.assertEqual(inst.cleanup_interval, 1100)
        self.assertEqual(inst.channel_timeout, 1200)
        self.assertEqual(inst.log_socket_errors, True)
        self.assertEqual(inst.max_request_header_size, 1300)
        self.assertEqual(inst.max_request_body_size, 1400)
        self.assertEqual(inst.expose_tracebacks, True)
        self.assertEqual(inst.asyncore_loop_timeout, 5)
        self.assertEqual(inst.asyncore_use_poll, True)
        self.assertEqual(inst.ident, 'abc')
        self.assertEqual(inst.unix_socket_perms, 0o777)
        self.assertEqual(inst.url_prefix, '/foo')
        self.assertEqual(inst.ipv4, True)
        self.assertEqual(inst.ipv6, False)

        bind_pairs = [
            sockaddr[:2]
            for (family, _, _, sockaddr) in inst.listen
            if family == socket.AF_INET
        ]

        # On Travis, somehow we start listening to two sockets when resolving
        # localhost...
        self.assertEqual(('127.0.0.1', 8080), bind_pairs[0])

    def test_goodvar_listen(self):
        inst = self._makeOne(listen='127.0.0.1')

        bind_pairs = [(host, port) for (_, _, _, (host, port)) in inst.listen]

        self.assertEqual(bind_pairs, [('127.0.0.1', 8080)])

    def test_default_listen(self):
        inst = self._makeOne()

        bind_pairs = [(host, port) for (_, _, _, (host, port)) in inst.listen]

        self.assertEqual(bind_pairs, [('0.0.0.0', 8080)])

    def test_multiple_listen(self):
        inst = self._makeOne(listen='127.0.0.1:9090 127.0.0.1:8080')

        bind_pairs = [sockaddr[:2] for (_, _, _, sockaddr) in inst.listen]

        self.assertEqual(bind_pairs,
                         [('127.0.0.1', 9090),
                          ('127.0.0.1', 8080)])

    def test_wildcard_listen(self):
        inst = self._makeOne(listen='*:8080')

        bind_pairs = [sockaddr[:2] for (_, _, _, sockaddr) in inst.listen]

        self.assertTrue(len(bind_pairs) >= 1)

    def test_ipv6_no_port(self): # pragma: nocover
        if not self._hasIPv6():
            return

        inst = self._makeOne(listen='[::1]')

        bind_pairs = [sockaddr[:2] for (_, _, _, sockaddr) in inst.listen]

        self.assertEqual(bind_pairs, [('::1', 8080)])

    def test_bad_port(self):
        self.assertRaises(ValueError, self._makeOne, listen='127.0.0.1:test')

    def test_service_port(self):
        if WIN and PY2: # pragma: no cover
            # On Windows and Python 2 this is broken, so we raise a ValueError
            self.assertRaises(
                ValueError,
                self._makeOne,
                listen='127.0.0.1:http',
            )
            return

        inst = self._makeOne(listen='127.0.0.1:http 0.0.0.0:https')

        bind_pairs = [sockaddr[:2] for (_, _, _, sockaddr) in inst.listen]

        self.assertEqual(bind_pairs, [('127.0.0.1', 80), ('0.0.0.0', 443)])

    def test_dont_mix_host_port_listen(self):
        self.assertRaises(
            ValueError,
            self._makeOne,
            host='localhost',
            port='8080',
            listen='127.0.0.1:8080',
        )

    def test_good_sockets(self):
        sockets = [
            socket.socket(socket.AF_INET6, socket.SOCK_STREAM),
            socket.socket(socket.AF_INET, socket.SOCK_STREAM)]
        inst = self._makeOne(sockets=sockets)
        self.assertEqual(inst.sockets, sockets)
        sockets[0].close()
        sockets[1].close()

    def test_dont_mix_sockets_and_listen(self):
        sockets = [socket.socket(socket.AF_INET, socket.SOCK_STREAM)]
        self.assertRaises(
            ValueError,
            self._makeOne,
            listen='127.0.0.1:8080',
            sockets=sockets)
        sockets[0].close()

    def test_dont_mix_sockets_and_host_port(self):
        sockets = [socket.socket(socket.AF_INET, socket.SOCK_STREAM)]
        self.assertRaises(
            ValueError,
            self._makeOne,
            host='localhost',
            port='8080',
            sockets=sockets)
        sockets[0].close()

    def test_dont_mix_sockets_and_unix_socket(self):
        sockets = [socket.socket(socket.AF_INET, socket.SOCK_STREAM)]
        self.assertRaises(
            ValueError,
            self._makeOne,
            unix_socket='./tmp/test',
            sockets=sockets)
        sockets[0].close()

    def test_dont_mix_unix_socket_and_host_port(self):
        self.assertRaises(
            ValueError,
            self._makeOne,
            unix_socket='./tmp/test',
            host='localhost',
            port='8080')

    def test_dont_mix_unix_socket_and_listen(self):
        self.assertRaises(
            ValueError,
            self._makeOne,
            unix_socket='./tmp/test',
            listen='127.0.0.1:8080')

    def test_dont_use_unsupported_socket_types(self):
        sockets = [socket.socket(socket.AF_INET, socket.SOCK_DGRAM)]
        self.assertRaises(
            ValueError,
            self._makeOne,
            sockets=sockets)
        sockets[0].close()

    def test_dont_mix_forwarded_with_x_forwarded(self):
        with self.assertRaises(ValueError) as cm:
            self._makeOne(trusted_proxy='localhost', trusted_proxy_headers={'forwarded', 'x-forwarded-for'})

        self.assertIn('The Forwarded proxy header', str(cm.exception))

    def test_unknown_trusted_proxy_header(self):
        with self.assertRaises(ValueError) as cm:
            self._makeOne(trusted_proxy='localhost', trusted_proxy_headers={'forwarded', 'x-forwarded-unknown'})

        self.assertIn(
            'unknown trusted_proxy_headers value (x-forwarded-unknown)',
            str(cm.exception)
        )

    def test_trusted_proxy_count_no_trusted_proxy(self):
        with self.assertRaises(ValueError) as cm:
            self._makeOne(trusted_proxy_count=1)

        self.assertIn(
            'trusted_proxy_count has no meaning',
            str(cm.exception)
        )

    def test_trusted_proxy_headers_no_trusted_proxy(self):
        with self.assertRaises(ValueError) as cm:
            self._makeOne(trusted_proxy_headers={'forwarded'})

        self.assertIn(
            'trusted_proxy_headers has no meaning',
            str(cm.exception)
        )

    def test_trusted_proxy_headers_string_list(self):
        inst = self._makeOne(trusted_proxy='localhost', trusted_proxy_headers='x-forwarded-for x-forwarded-by')
        self.assertEqual(inst.trusted_proxy_headers, {'x-forwarded-for', 'x-forwarded-by'})

    def test_trusted_proxy_headers_string_list_newlines(self):
        inst = self._makeOne(trusted_proxy='localhost', trusted_proxy_headers='x-forwarded-for\nx-forwarded-by\nx-forwarded-host')
        self.assertEqual(inst.trusted_proxy_headers, {'x-forwarded-for', 'x-forwarded-by', 'x-forwarded-host'})

    def test_no_trusted_proxy_headers_trusted_proxy(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.resetwarnings()
            warnings.simplefilter("always")
            self._makeOne(trusted_proxy='localhost')

            self.assertGreaterEqual(len(w), 1)
            self.assertTrue(issubclass(w[0].category, DeprecationWarning))
            self.assertIn("Implicitly trusting X-Forwarded-Proto", str(w[0]))

    def test_clear_untrusted_proxy_headers(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.resetwarnings()
            warnings.simplefilter("always")
            self._makeOne(trusted_proxy='localhost', trusted_proxy_headers={'x-forwarded-for'})

            self.assertGreaterEqual(len(w), 1)
            self.assertTrue(issubclass(w[0].category, DeprecationWarning))
            self.assertIn("clear_untrusted_proxy_headers will be set to True", str(w[0]))

    def test_badvar(self):
        self.assertRaises(ValueError, self._makeOne, nope=True)

    def test_ipv4_disabled(self):
        self.assertRaises(ValueError, self._makeOne, ipv4=False, listen="127.0.0.1:8080")

    def test_ipv6_disabled(self):
        self.assertRaises(ValueError, self._makeOne, ipv6=False, listen="[::]:8080")

    def test_server_header_removable(self):
        inst = self._makeOne(ident=None)
        self.assertEqual(inst.ident, None)

        inst = self._makeOne(ident='')
        self.assertEqual(inst.ident, None)

        inst = self._makeOne(ident='specific_header')
        self.assertEqual(inst.ident, 'specific_header')


class TestCLI(unittest.TestCase):

    def parse(self, argv):
        from waitress.adjustments import Adjustments
        return Adjustments.parse_args(argv)

    def test_noargs(self):
        opts, args = self.parse([])
        self.assertDictEqual(opts, {'call': False, 'help': False})
        self.assertSequenceEqual(args, [])

    def test_help(self):
        opts, args = self.parse(['--help'])
        self.assertDictEqual(opts, {'call': False, 'help': True})
        self.assertSequenceEqual(args, [])

    def test_call(self):
        opts, args = self.parse(['--call'])
        self.assertDictEqual(opts, {'call': True, 'help': False})
        self.assertSequenceEqual(args, [])

    def test_both(self):
        opts, args = self.parse(['--call', '--help'])
        self.assertDictEqual(opts, {'call': True, 'help': True})
        self.assertSequenceEqual(args, [])

    def test_positive_boolean(self):
        opts, args = self.parse(['--expose-tracebacks'])
        self.assertDictContainsSubset({'expose_tracebacks': 'true'}, opts)
        self.assertSequenceEqual(args, [])

    def test_negative_boolean(self):
        opts, args = self.parse(['--no-expose-tracebacks'])
        self.assertDictContainsSubset({'expose_tracebacks': 'false'}, opts)
        self.assertSequenceEqual(args, [])

    def test_cast_params(self):
        opts, args = self.parse([
            '--host=localhost',
            '--port=80',
            '--unix-socket-perms=777'
        ])
        self.assertDictContainsSubset({
            'host': 'localhost',
            'port': '80',
            'unix_socket_perms': '777',
        }, opts)
        self.assertSequenceEqual(args, [])

    def test_listen_params(self):
        opts, args = self.parse([
            '--listen=test:80',
            ])

        self.assertDictContainsSubset({
            'listen': ' test:80'
            }, opts)
        self.assertSequenceEqual(args, [])

    def test_multiple_listen_params(self):
        opts, args = self.parse([
            '--listen=test:80',
            '--listen=test:8080',
            ])

        self.assertDictContainsSubset({
            'listen': ' test:80 test:8080'
            }, opts)
        self.assertSequenceEqual(args, [])

    def test_bad_param(self):
        import getopt
        self.assertRaises(getopt.GetoptError, self.parse, ['--no-host'])


if hasattr(socket, 'AF_UNIX'):
    class TestUnixSocket(unittest.TestCase):
        def _makeOne(self, **kw):
            from waitress.adjustments import Adjustments
            return Adjustments(**kw)

        def test_dont_mix_internet_and_unix_sockets(self):
            sockets = [
                socket.socket(socket.AF_INET, socket.SOCK_STREAM),
                socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)]
            self.assertRaises(
                ValueError,
                self._makeOne,
                sockets=sockets)
            sockets[0].close()
            sockets[1].close()



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import socket
import threading
import time
import traceback

from waitress.buffers import (
    OverflowableBuffer,
    ReadOnlyFileBasedBuffer,
)

from waitress.parser import HTTPRequestParser

from waitress.task import (
    ErrorTask,
    WSGITask,
)

from waitress.utilities import InternalServerError

from . import wasyncore

class HTTPChannel(wasyncore.dispatcher, object):
    """
    Setting self.requests = [somerequest] prevents more requests from being
    received until the out buffers have been flushed.

    Setting self.requests = [] allows more requests to be received.
    """

    task_class = WSGITask
    error_task_class = ErrorTask
    parser_class = HTTPRequestParser

    request = None               # A request parser instance
    last_activity = 0            # Time of last activity
    will_close = False           # set to True to close the socket.
    close_when_flushed = False   # set to True to close the socket when flushed
    requests = ()                # currently pending requests
    sent_continue = False        # used as a latch after sending 100 continue
    force_flush = False          # indicates a need to flush the outbuf

    #
    # ASYNCHRONOUS METHODS (including __init__)
    #

    def __init__(
            self,
            server,
            sock,
            addr,
            adj,
            map=None,
            ):
        self.server = server
        self.adj = adj
        self.outbufs = [OverflowableBuffer(adj.outbuf_overflow)]
        self.creation_time = self.last_activity = time.time()

        # task_lock used to push/pop requests
        self.task_lock = threading.Lock()
        # outbuf_lock used to access any outbuf
        self.outbuf_lock = threading.Lock()

        wasyncore.dispatcher.__init__(self, sock, map=map)

        # Don't let wasyncore.dispatcher throttle self.addr on us.
        self.addr = addr

    def any_outbuf_has_data(self):
        for outbuf in self.outbufs:
            if bool(outbuf):
                return True
        return False

    def total_outbufs_len(self):
        # genexpr == more funccalls
        # use b.__len__ rather than len(b) FBO of not getting OverflowError
        # on Python 2
        return sum([b.__len__() for b in self.outbufs]) 

    def writable(self):
        # if there's data in the out buffer or we've been instructed to close
        # the channel (possibly by our server maintenance logic), run
        # handle_write
        return self.any_outbuf_has_data() or self.will_close

    def handle_write(self):
        # Precondition: there's data in the out buffer to be sent, or
        # there's a pending will_close request
        if not self.connected:
            # we dont want to close the channel twice
            return

        # try to flush any pending output
        if not self.requests:
            # 1. There are no running tasks, so we don't need to try to lock
            #    the outbuf before sending
            # 2. The data in the out buffer should be sent as soon as possible
            #    because it's either data left over from task output
            #    or a 100 Continue line sent within "received".
            flush = self._flush_some
        elif self.force_flush:
            # 1. There's a running task, so we need to try to lock
            #    the outbuf before sending
            # 2. This is the last chunk sent by the Nth of M tasks in a
            #    sequence on this channel, so flush it regardless of whether
            #    it's >= self.adj.send_bytes.  We need to do this now, or it
            #    won't get done.
            flush = self._flush_some_if_lockable
            self.force_flush = False
        elif (self.total_outbufs_len() >= self.adj.send_bytes):
            # 1. There's a running task, so we need to try to lock
            #    the outbuf before sending
            # 2. Only try to send if the data in the out buffer is larger
            #    than self.adj_bytes to avoid TCP fragmentation
            flush = self._flush_some_if_lockable
        else:
            # 1. There's not enough data in the out buffer to bother to send
            #    right now.
            flush = None

        if flush:
            try:
                flush()
            except socket.error:
                if self.adj.log_socket_errors:
                    self.logger.exception('Socket error')
                self.will_close = True
            except:
                self.logger.exception('Unexpected exception when flushing')
                self.will_close = True

        if self.close_when_flushed and not self.any_outbuf_has_data():
            self.close_when_flushed = False
            self.will_close = True

        if self.will_close:
            self.handle_close()

    def readable(self):
        # We might want to create a new task.  We can only do this if:
        # 1. We're not already about to close the connection.
        # 2. There's no already currently running task(s).
        # 3. There's no data in the output buffer that needs to be sent
        #    before we potentially create a new task.
        return not (self.will_close or self.requests or
                    self.any_outbuf_has_data())

    def handle_read(self):
        try:
            data = self.recv(self.adj.recv_bytes)
        except socket.error:
            if self.adj.log_socket_errors:
                self.logger.exception('Socket error')
            self.handle_close()
            return
        if data:
            self.last_activity = time.time()
            self.received(data)

    def received(self, data):
        """
        Receives input asynchronously and assigns one or more requests to the
        channel.
        """
        # Preconditions: there's no task(s) already running
        request = self.request
        requests = []

        if not data:
            return False

        while data:
            if request is None:
                request = self.parser_class(self.adj)
            n = request.received(data)
            if request.expect_continue and request.headers_finished:
                # guaranteed by parser to be a 1.1 request
                request.expect_continue = False
                if not self.sent_continue:
                    # there's no current task, so we don't need to try to
                    # lock the outbuf to append to it.
                    self.outbufs[-1].append(b'HTTP/1.1 100 Continue\r\n\r\n')
                    self.sent_continue = True
                    self._flush_some()
                    request.completed = False
            if request.completed:
                # The request (with the body) is ready to use.
                self.request = None
                if not request.empty:
                    requests.append(request)
                request = None
            else:
                self.request = request
            if n >= len(data):
                break
            data = data[n:]

        if requests:
            self.requests = requests
            self.server.add_task(self)

        return True

    def _flush_some_if_lockable(self):
        # Since our task may be appending to the outbuf, we try to acquire
        # the lock, but we don't block if we can't.
        locked = self.outbuf_lock.acquire(False)
        if locked:
            try:
                self._flush_some()
            finally:
                self.outbuf_lock.release()

    def _flush_some(self):
        # Send as much data as possible to our client

        sent = 0
        dobreak = False

        while True:
            outbuf = self.outbufs[0]
            # use outbuf.__len__ rather than len(outbuf) FBO of not getting
            # OverflowError on Python 2
            outbuflen = outbuf.__len__()
            if outbuflen <= 0:
                # self.outbufs[-1] must always be a writable outbuf
                if len(self.outbufs) > 1:
                    toclose = self.outbufs.pop(0)
                    try:
                        toclose.close()
                    except:
                        self.logger.exception(
                            'Unexpected error when closing an outbuf')
                    continue # pragma: no cover (coverage bug, it is hit)
                else:
                    if hasattr(outbuf, 'prune'):
                        outbuf.prune()
                    dobreak = True

            while outbuflen > 0:
                chunk = outbuf.get(self.adj.send_bytes)
                num_sent = self.send(chunk)
                if num_sent:
                    outbuf.skip(num_sent, True)
                    outbuflen -= num_sent
                    sent += num_sent
                else:
                    dobreak = True
                    break

            if dobreak:
                break

        if sent:
            self.last_activity = time.time()
            return True

        return False

    def handle_close(self):
        for outbuf in self.outbufs:
            try:
                outbuf.close()
            except:
                self.logger.exception(
                    'Unknown exception while trying to close outbuf')
        self.connected = False
        wasyncore.dispatcher.close(self)

    def add_channel(self, map=None):
        """See wasyncore.dispatcher

        This hook keeps track of opened channels.
        """
        wasyncore.dispatcher.add_channel(self, map)
        self.server.active_channels[self._fileno] = self

    def del_channel(self, map=None):
        """See wasyncore.dispatcher

        This hook keeps track of closed channels.
        """
        fd = self._fileno # next line sets this to None
        wasyncore.dispatcher.del_channel(self, map)
        ac = self.server.active_channels
        if fd in ac:
            del ac[fd]

    #
    # SYNCHRONOUS METHODS
    #

    def write_soon(self, data):
        if data:
            # the async mainloop might be popping data off outbuf; we can
            # block here waiting for it because we're in a task thread
            with self.outbuf_lock:
                if data.__class__ is ReadOnlyFileBasedBuffer:
                    # they used wsgi.file_wrapper
                    self.outbufs.append(data)
                    nextbuf = OverflowableBuffer(self.adj.outbuf_overflow)
                    self.outbufs.append(nextbuf)
                else:
                    self.outbufs[-1].append(data)
            # XXX We might eventually need to pull the trigger here (to
            # instruct select to stop blocking), but it slows things down so
            # much that I'll hold off for now; "server push" on otherwise
            # unbusy systems may suffer.
            return len(data)
        return 0

    def service(self):
        """Execute all pending requests """
        with self.task_lock:
            while self.requests:
                request = self.requests[0]
                if request.error:
                    task = self.error_task_class(self, request)
                else:
                    task = self.task_class(self, request)
                try:
                    task.service()
                except:
                    self.logger.exception('Exception when serving %s' %
                                          task.request.path)
                    if not task.wrote_header:
                        if self.adj.expose_tracebacks:
                            body = traceback.format_exc()
                        else:
                            body = ('The server encountered an unexpected '
                                    'internal server error')
                        req_version = request.version
                        req_headers = request.headers
                        request = self.parser_class(self.adj)
                        request.error = InternalServerError(body)
                        # copy some original request attributes to fulfill
                        # HTTP 1.1 requirements
                        request.version = req_version
                        try:
                            request.headers['CONNECTION'] = req_headers[
                                'CONNECTION']
                        except KeyError:
                            pass
                        task = self.error_task_class(self, request)
                        task.service() # must not fail
                    else:
                        task.close_on_finish = True
                # we cannot allow self.requests to drop to empty til
                # here; otherwise the mainloop gets confused
                if task.close_on_finish:
                    self.close_when_flushed = True
                    for request in self.requests:
                        request.close()
                    self.requests = []
                else:
                    request = self.requests.pop(0)
                    request.close()

        self.force_flush = True
        self.server.pull_trigger()
        self.last_activity = time.time()

    def cancel(self):
        """ Cancels all pending requests """
        self.force_flush = True
        self.last_activity = time.time()
        self.requests = []

    def defer(self):
        pass



from waitress.server import create_server
import logging

def serve(app, **kw):
    _server = kw.pop('_server', create_server) # test shim
    _quiet = kw.pop('_quiet', False) # test shim
    _profile = kw.pop('_profile', False) # test shim
    if not _quiet: # pragma: no cover
        # idempotent if logging has already been set up
        logging.basicConfig()
    server = _server(app, **kw)
    if not _quiet: # pragma: no cover
        server.print_listen('Serving on http://{}:{}')
    if _profile: # pragma: no cover
        profile('server.run()', globals(), locals(), (), False)
    else:
        server.run()

def serve_paste(app, global_conf, **kw):
    serve(app, **kw)
    return 0

def profile(cmd, globals, locals, sort_order, callers): # pragma: no cover
    # runs a command under the profiler and print profiling output at shutdown
    import os
    import profile
    import pstats
    import tempfile
    fd, fn = tempfile.mkstemp()
    try:
        profile.runctx(cmd, globals, locals, fn)
        stats = pstats.Stats(fn)
        stats.strip_dirs()
        # calls,time,cumulative and cumulative,calls,time are useful
        stats.sort_stats(*(sort_order or ('cumulative', 'calls', 'time')))
        if callers:
            stats.print_callers(.3)
        else:
            stats.print_stats(.3)
    finally:
        os.remove(fn)
